#!/usr/bin/python3
# coding: utf-8

import os
import struct, sys, socket, time #,json
import configparser,  csv
import traceback
from weblib import *

exponentGlobal = []
nameGlobal = ""
typeGlobal = ""
primaryAddress = 0      # Primary address.
mbusini = configparser.RawConfigParser()
mbusini.read("/var/www1/MbusOPC.ini" ,encoding="ISO-8859-1")

# List of days in week
DaysInWeek = {
    0x00:"Not specified", 0x01:"Monday", 0x02:"Tuesday", 0x03:"Wednesday",
    0x04:"Thursday", 0x05:"Friday", 0x06:"Saturday", 0x07:"Sunday"
}

def MfctToStr(mfct):
    if mfct == 0xFFFF:
        return ""
    strng  = "%c"%(((mfct>>10)&0x1F)+64)
    strng += "%c"%(((mfct>>5)&0x1F)+64)
    strng += "%c"%((mfct&0x1F)+64) # Manufacturer number
    return strng

# Combinable (orthogonal) VIFE-codes.
def getCombinableVIFE(vife):
    #medium = ""
    exponent = 0
    e = False
    physicalQuantity = ""
    templateVIB = ""
    exponent = ""
    a=mbusini.get("VIFE", "%02X"%(vife&0x7F))
    a=a.split(',')
    if len(a[0]) > 0:
        exponent = a[0]
        unit = a[1]
        physicalQuantity = a[3]
        templateVIB = a[5]
    ret = ""
    if e:
        ret = "{:s} {:4s} {:s}".format(physicalQuantity, unit, exponent)
    else:
        ret = "{:s} {:4s}".format(physicalQuantity, unit)
    return ret

# Plain-text VIF.
def getPlainASCII(length, asci):
    asci = bytearray(asci)
    asci.reverse()
    #print("ascii reverse", asci)
    return "".join(chr(i) for i in asci)
    # Kolla Elvaco prim.adr 19.

# Plain-text VIF. Replacing dots and spaces.
def getPlainASCII_1(length, asci):
    asci = bytearray(asci)
    asci.reverse()
    #print("ascii reverse", ascii)
    retval = "".join(chr(i) for i in asci)
    retval = retval.replace(". ","_").replace(".","_").replace(" ","_")
    return retval
    # Kolla Elvaco prim.adr 19.

# Third extension VIFE-codes.
def getVIFE_EF(vife):
    #print("VIFE = %02X"%(vife))
    # Reserved for future use. Not implemented in EN13757-3.
    return ""

def getVIFE_FF(vife,  opcname):
    #print("VIFE = %02X"%(vife))
    unit = "Not implemented VIFE FF"
    physicalQuantity = ""
    templateVIB = "VIFE_FF"
    exponent = ""
    try:
        a=mbusini.get(opcname, "%02X"%(vife&0x7F))
        a=a.split(',')
        if len(a) > 0:
            if bool(a[2]) and  len(a[0])>0:
                exponent = a[0]
            unit = a[1]
            physicalQuantity = a[3]
            templateVIB=a[5]
    except:
        pass
    isDateTime = False
    return "{:s} {:4s}".format(physicalQuantity, unit), exponent, isDateTime,  templateVIB

# Second extension VIFE-codes.
def getVIFE_FD(vife):
    #print("VIFE = %02X"%(vife))
    unit = "Not implemented VIFE FD"
    physicalQuantity = ""
    exponent = ""
    templateVIB = ""
    try: #If this fails, something is missing from MBus opc.ini
        a=mbusini.get("VIFE_FD", "%02X"%(vife&0x7F))
        a=a.split(',')
        if len(a) > 0:
            if bool(a[2]) and  len(a[0])>0:
                exponent = a[0]
            unit = a[1]
            physicalQuantity = a[3]
            templateVIB=a[5]
    except:
        pass
    isDateTime = False
    return "{:s} {:4s}".format(physicalQuantity, unit), exponent, isDateTime, templateVIB

# First extension VIFE-codes.
def getVIFE_FB(vife):
    #print("VIFE = %02X"%(vife))
    unit = "Not implemented VIFE FB"
    physicalQuantity = ""
    exponent = ""
    a=mbusini.get("VIFE_FB", "%02X"%(vife&0x7F))
    a=a.split(',')
    templateVIB=""
    try: #If this fails, something is missing from MBus opc.ini
        if len(a) > 0:
            exponent = a[0]
            unit = a[1]
            physicalQuantity = a[3]
            templateVIB=a[5]
    except:
        pass
    isDateTime = False
    return "{:s} {:4s}".format(physicalQuantity, unit), exponent, isDateTime,  templateVIB

# Primary VIF-codes.
def getVIF(vif):
    #print("VIF = %02X"%(vif))
    unit = "Not implemented VIF %02X"%vif
    templateVIB= "VIF%02X"%vif
    physicalQuantity = ""
    exponent=""
    isDateTime = False
    if vif == 0x6c or vif==0x6d:
        isDateTime = True
    a=mbusini.get("VIF", "%02X"%(vif&0x7F))
    a=a.split(',')
    if len(a) > 0:
        if len(a[0])>0:
            exponent = a[0]
        unit = a[1]
        physicalQuantity = a[3]
        templateVIB=a[5]
    return "{:s} {:4s}".format(physicalQuantity, unit), exponent, isDateTime,  templateVIB

def printhex(data):
    tstr = ""
    for j in range(len(data)):
        if j%16 == 0 and j != 0:
            tstr += "\n"
        tstr += "%02X "%data[j]
    print(tstr)
    print("")

def CalcBcc(array):
    cc=0
    for i in range(len(array)):
        cc += array[i]
    return cc&0xFF

def bcdtoval(arr):
    # Kolla: Prim.Adr 4, objekt 35 innehåller 0xFF. Hur ska det tolkas?
    retval=0
    mul=1
    for i in arr:
        j = i
        retval = retval+(j&0x0f)*mul
        mul = mul * 10;
        retval = retval+((j&0xf0)>>4)*mul
        mul = mul * 10;
    return retval

def SecondaryAddress(data):
    data = bytearray(data)
    id = data[0:4]
    id.reverse()
    id = "".join("%02X" % b for b in id)
    manu = data[4:6]
    manu.reverse()
    manu = "".join("%02X" % b for b in manu)
    ver = data[6]
    med = data[7]
    
    tmp = struct.unpack("<H",data[4:6])[0]
    name = MfctToStr(tmp)
    lansen_exception_medium = [0x1D, 0x1E, 0x2B]
    if name == "LAS" and med in lansen_exception_medium:
        a=mbusini.get("MEDIA_LANSEN", "%02X"%(med%0x7F))
    else:
        a=mbusini.get("MEDIA_VARIABLE", "%02X"%(med%0x7F))
    a=a.split(',')
    meter_type = "Reserved"
    print("a",a)
    if len(a[0]) > 0:
        meter_type = a[0]
    return id, manu, "%02X"%(ver), "%02X"%(med), name, meter_type

def Status(value):
    result = ""

    # Status for bit 0 and 1.
    if (value & 0x03):
        bit1_2 = value & 0x03
        result = "Application busy" if bit1_2 == 1 else "Any application error" if bit1_2 == 2 else "Abnormal condition / alarm"

    # Status for bit 2.
    if (value & 0x04) == 0x04:
        result += ", " if len(result) > 0 else result
        result += "Power low"

    # Status for bit 3.
    if (value & 0x08) == 0x08:
        result += ", " if len(result) > 0 else result
        result += "Permanent error"

    # Status for bit 4.
    if (value & 0x10) == 0x010:
        result += ", " if len(result) > 0 else result
        result += "Temporary error"

    # Status for bit 5.
    if (value & 0x20) == 0x20:
        result += ", " if len(result) > 0 else result
        result += "Manufacturer bit1 set"

    # Status for bit 6.
    if (value & 0x40) == 0x40:
        result += ", " if len(result) > 0 else result
        result += "Manufacturer bit1 set"

    # Status for bit 7.
    if (value & 0x80) == 0x80:
        result += ", " if len(result) > 0 else result
        result += "Manufacturer bit1 set"

    return result

# Decode CP16 from INT16 (Type G) to YY-MM-DD.
def DecodeCP16(value):
    # Prim.adr 42 objekt 4. Prim.adr 43 objekt 4, 9, 11.
    y1 = (value & 0xF000) >> 8
    y2 = value & 0x00E0
    mon = (value & 0x0F00) >> 8
    day = value & 0x001F
    year = (y1 >> 1) | (y2 >> 5)
    return "%02d-%02d-%02d"%(year, mon, day)

# Decode CP24 from INT24 (Type J) to HH:MM:SS.
def DecodeCP24(value):
    sec = value & 0x00003F
    min = (value & 0x003F00) >> 8
    hour = (value & 0x1F0000) >> 16
    return "%02d:%02d:%02d"%(hour, min, sec)

# Decode CP32 from INT32 (Type F) to YY-MM-DD HH:MM.
def DecodeCP32(value):
    # Prim.adr 25, objekt 20.
    min  =  value & 0x0000003F
    hour = (value & 0x00001F00) >> 8
    day  = (value & 0x001F0000) >> 16
    mon  = (value & 0x0F000000) >> 24
    y1   = (value & 0xF0000000) >> 24
    y2   = (value & 0x00E00000) >> 16
    #hy   = (value & 0x00006000) >> 8 # Hundred years
    year = (y1 >> 1) | (y2 >> 5)
    return "%02d-%02d-%02d %02d:%02d"%(year, mon, day, hour, min)

# Decode CP48 from INT48 (Type I) to YY-MM-DD HH:MM:SS, day, week, period.
def DecodeCP48(value):
    #print("JJJJJ")
    '''
         48    41   40    33   32    25   24    17   16     9   8      1
         47    40   39    32   31    24   23    16   15     8   7      0
    DT = 00000000 | 00000000 | 00000000 | 00000000 | 00000000 | 00000000
         --______   ----____   ---_____   ---_____   -_------   _-______
         ds Week	 y1 mon     y2 day    dow hour   ih  min    lt  sec
    '''
    #ds 	    = (value & 0xC00000000000) >> 40   # Daylight saving deviation (hour).
    week    = (value & 0x3F0000000000) >> 40
    y1	    = (value & 0x00F000000000) >> 33
    mon	    = (value & 0x000F00000000) >> 32
    y2	    = (value & 0x0000E0000000) >> 29
    day	    = (value & 0x00001F000000) >> 24
    dow	    = (value & 0x000000E00000) >> 21   # Day of week.
    hour	    = (value & 0x0000001F0000) >> 16
    #i	        = (value & 0x000000008000) >> 8    # Time invalid.
    #h	    = (value & 0x000000004000) >> 8    # Daylight saving deviation (hour).
    min	    = (value & 0x000000003F00) >> 8
    #l	    = (value & 0x000000000080)         # Leap year.
    #per     = (value & 0x000000000040)         # Time during daylight savings, period.
    sec	    = (value & 0x00000000003F)

    year = y1 + y2
    dayInWeek = DaysInWeek[dow]                         # Day in week.
    period = "winter"
    if period != 0x00:
        period = "summer"   # Winter or summer time period.

    strng =  "%02d-%02d-%02d %02d:%02d:%02d %s %02d %s"%(year, mon, day, hour, min, sec, dayInWeek, week, period)
    return strng

# Decode a BCD to DateTime. Unsupported coding of date and time, but is common for ABB and BMeter.
def DecodeBCD12(value):
    dateTime = "%012d"%(value)
    year = dateTime[0:2]
    mon =  dateTime[2:4]
    day =  dateTime[4:6]
    hour = dateTime[6:8]
    min =  dateTime[8:10]
    sec =  dateTime[10:12]
    return "%s-%s-%s %s:%s:%s"%(year, mon, day, hour, min, sec)

# Decode CP16, CP24, CP32 and CP48.
def DecodeDateTime(dataType, value):
    if dataType == 0x02: # INT16 = CP16.
        vib = "Date YY-MM-DD"
        return DecodeCP16(value), vib
    elif dataType == 0x03: # INT24 = CP24.
        vib = "Time HH:MM:SS"
        return DecodeCP24(value), vib
    elif dataType == 0x04: # INT32 = CP32.
        vib = "Date & time YY-MM-DD HH:MM"
        return DecodeCP32(value), vib
    elif dataType == 0x06: # INT48 = CP48.
        vib = "Date & time YY-MM-DD HH:MM:SS, day, week, period"
        return DecodeCP48(value), vib
    elif dataType == 0x0E: # BCD12 = Unsupported date and time.
        vib = "Date & time YY-MM-DD HH:MM:SS"
        return DecodeBCD12(value), vib

def loopVIFE(vif, recordLength, arr, vife):
    if vif & 0x80: # If VIFE is present, check all VIFE.
        for i in range(10):
            vife=arr[recordLength]              # Get VIFE.
            #print("VIFE = 0x%02X"%vife)
            #if hasManufacturerSpecificVIB == False:
            #    comb = getCombinableVIFE(arr[recordLength+1])  # Combinable (orthogonal) VIFE-codes.
            #    print(comb)
            #    vib += comb
            recordLength += 1     # Point to next byte.
            if (vife&0x80)==0: # Check if extension bit is set in current VIFE.
                break
    return recordLength, vife

# Decode object
def decode_field(arr,  opcname, telegramCounter):
    dif=arr[0]      # Get DIF.
    #print("\tDIF\t\t= 0x%02X"%dif)
    value=None
    dataTypeStr=""              # String representation of the data type.
    recordLength=1              # Point to second byte in arr, may be DIFE or VIF.
    more_telegrams=False
    mfct_data=False
    functionFields=["Instantaneous value","Minimum value", "Maximum value", "Value during error state"]
    function=functionFields[(dif>>4)&0x03]
    dataType = dif & 0x0F       # Get data type.
    vib = ""
    isDateTime = False
    exponent = ""
    abb_version_data = ""
    abb_version_object = False
    #hasManufacturerSpecificVIB = False

    try:
        # === Parse DIB ===========================================================
        if dataType == 0x0F: # DIF = MDH.
            value = "None"
            mdh = (dif>>4)&0x07  # Get MDH.
            dataTypeStr = "0x%sF"%(mdh)
            if mdh==0: # MDH = 0x0F.
                function="Manufacturer specific, last telegram"
                mfct_data=True
                recordLength = len(arr) - 2
                value = ""
                for i in range(recordLength):
                    value += "%02X "%struct.unpack("B", arr[i:i+1])[0]
                vib = "Manufacturer specific, last telegram"
            elif mdh==1: # MDH = 0x1F.
                function="Manufacturer specific, more telegrams"
                mfct_data=True
                more_telegrams=True
                recordLength = len(arr) - 2
                value = ""
                print("value")
                for i in range(recordLength):
                    value += "%02X "%struct.unpack("B", arr[i:i+1])[0]
                    print(value)
                vib = "Manufacturer specific, more telegrams"
            elif mdh==2: # MDH = 0x2F.
                function="Idle filler, following byte is a DIF"
                vib = "Idle filler"
                value = None
            elif mdh==7: # MDH = 0x7F.
                function="Global readout request"
                vib = "Global readout request"
            else:
                function="Reserved"
                vib = "Reserved"

            return_map = {
                "value": value,
                "dataTypeStr": dataTypeStr,
                "recordLength": recordLength,
                "function": function,
                "mfct_data": mfct_data,
                "more_telegrams": more_telegrams,
                "vib": vib,
                "exponent": exponent,
                "templateVIB": "mfct",
                "abb_version_data": abb_version_data
            }
            return return_map

        elif dif & 0x80: # If DIFE is present, check all DIFE.
            for i in range(10):
                dife=arr[recordLength]              # Get DIFE.
                #print("\tDIFE\t\t= 0x%02X"%dife)
                recordLength += 1     # Point to next byte.
                # Check if extension bit is set in current DIFE.
                if (dife&0x80)==0:
                    break

    

        # === Parse VIB ===========================================================
        vif=arr[recordLength]  # Get VIB
        
        if (vif == 0xFF and 
            arr[recordLength+1] == 0xAA and 
            arr[recordLength+2] == 0x00 and 
            dataType == 0x0D and
            telegramCounter == 1):
            
            abb_version_object = True

        vife = 0x00
        templateVIB = ""
        if vif < 0xFB and vif != 0x7F and vif != 0x7C:  # Primary VIF-codes.
            vib, exponent,  isDateTime,  templateVIB = getVIF(vif)
        elif vif == 0xFB: # First extension of VIF-codes.
            vife = arr[recordLength + 1]
            vib, exponent,  isDateTime,  templateVIB = getVIFE_FB(vife)
        elif vif == 0x7C:   # Plain ASCII.
            # VIF LEN ASCII_1 ... ASCII_N DATA         (No VIFE present).
            recordLength += 1
            length = arr[recordLength]
            print("7C length", length)
            recordLength += 1
            # Point first byte after length and include as many bytes as specified in length.
            ascii = arr[recordLength:recordLength+length]
            #vib, exponent = getPlainASCII(length, ascii)
            vib = getPlainASCII(length, ascii)
            templateVIB = getPlainASCII_1(length, asci)
            recordLength += length
        elif vif == 0xFC:
            # Note:
            # The VIFE will follow after the length and data.
            # Ex1: VIF LEN ASCII_1 ... ASCII_N VIFE DATA    (VIFE present).
            # Ex2: VIF LEN ASCII_1 ... ASCII_N DATA         (No VIFE present).
            length = arr[recordLength + 1]
            # Point first byte after length and include as many bytes as specified in length.
            ascii = arr[recordLength+2:recordLength+2+arr[recordLength+1]]
            #vib, exponent = getPlainASCII(length, ascii)
            vib = getPlainASCII(length, ascii)
            recordLength += length + 1
        elif vif == 0xFD: # Second extension of VIF-codes.
            vife = arr[recordLength + 1]
            vib, exponent,  isDateTime,  templateVIB = getVIFE_FD(vife)
        elif vif == 0xEF: # (Reserved) Third extension VIF-codes.
            vife = arr[recordLength + 1]
            vib, exponent,  isDateTime,  templateVIB = getVIFE_EF(vife)
        elif vif == 0x7E or vif == 0xFE: # Any VIF.
            vib = "Any VIF"
        elif vif == 0x7F:
            vib, exponent,  isDateTime,  templateVIB = getVIF(vif)
        elif vif == 0xFF: # Manufacturer specific.
            #vib = "Manufacturer specific"
            vife = arr[recordLength + 1]
            vib, exponent,  isDateTime,  templateVIB = getVIFE_FF(vife, opcname)
            #hasManufacturerSpecificVIB = True
        else: # Unknown VIF.
            vib = "Unknown"

        if vif != 0x7C:
            recordLength +=  1 # Point to next byte (VIFE or start of data area).

        recordLength, vife = loopVIFE(vif, recordLength, arr, vife)

        # === Parse DATA ==================================================
        # No data.
        #printhex(arr)
        if dataType==0:
            dataTypeStr="No Data"
        elif dataType==0x01: # INT8.
            dataTypeStr="INT8"
            value=struct.unpack('<b',arr[recordLength:recordLength+1])
            value = value[0]
            recordLength += 1
        elif dataType==0x02: # INT16.
            #printhex(arr[recordLength:])
            dataTypeStr="INT16"
            value=struct.unpack('<h',arr[recordLength:recordLength+2])
            value = value[0]
            recordLength +=  2
        elif dataType==0x03: # INT24.
            dataTypeStr="INT24"
            value = struct.unpack('<i', b'\x00' + arr[recordLength:recordLength + 3])
            value = value[0] >> 8
            recordLength += 3
        elif dataType==0x04: # INT32.
            #printhex(arr)
            #printhex(arr[recordLength:])
            dataTypeStr="INT32"
            value=struct.unpack('<i',arr[recordLength:recordLength+4])
            value = value[0]
            recordLength += 4
        elif dataType==0x05: # FLOAT.
            dataTypeStr="REAL32"
            value=struct.unpack('<f',arr[recordLength:recordLength+4])
            value = value[0]
            recordLength += 4
        elif dataType==0x06: # INT48.
            #print("HHHH INT48")
            #printhex(arr[recordLength:recordLength+6])
            dataTypeStr="INT48"
            value=struct.unpack('<q',b'\x00\x00' + arr[recordLength:recordLength+6])
            value = value[0] >> 16
            #print("IIII")
            recordLength += 6
        elif dataType==0x07:  # INT64.
            dataTypeStr="INT64"
            value=struct.unpack('<q',arr[recordLength:recordLength+8])
            value = value[0]
            recordLength += 8
        elif dataType==0x08: # Selection for readout.
            #dataTypeStr="Selection for Readout"
            dataTypeStr="SEL"
            recordLength = recordLength + 8
        elif dataType==0x09:  # BCD2.
            dataTypeStr="BCD2"
            value=bcdtoval(arr[recordLength:recordLength+1])
            recordLength += 1
        elif dataType==0x0A: # BCD4.
            dataTypeStr="BCD4"
            value=bcdtoval(arr[recordLength:recordLength+2])
            recordLength += 2
        elif dataType==0x0B: # BCD6.
            dataTypeStr="BCD6"
            value=bcdtoval(arr[recordLength:recordLength+3])
            recordLength = recordLength + 3
        elif dataType==0x0C: # BCD8.
            dataTypeStr="BCD8"
            value=bcdtoval(arr[recordLength:recordLength+4])
            recordLength += 4
        elif dataType==0x0D:  # Variable length.
            LVAR=arr[recordLength]
            dataLength = 0
            recordLength += 1
            dataTypeStr = "LVAR"
            # There is a bug in Browse. The LVAR container for wireless is not shown correctly.
            if arr[recordLength-3] == 0xFD and arr[recordLength-2] == 0x3B: # Wireless container.
                dataLength = LVAR
                data = arr[recordLength:recordLength+dataLength]
                value = "".join("%02X "%(i) for i in data)            
            elif LVAR <= 0xBF:  # 8-bit text (ISO 8859-1).
                #dataTypeStr="8-bit text (ISO 8859-1)"
                dataLength = LVAR
                print("dataLength in 0x0D", dataLength)
                #dataLength = arr[recordLength]
                #data = arr[recordLength:recordLength+dataLength]
                #value = "".join(chr(i) for i in data[::-1])
                value = "" #Change temporarily to HEX output. XXXXXXXXXXXXXXXXXXX. Ugly fix to a bug Must be fixed
                #print("LVAR", LVAR)
                #for i in range(dataLength):
                    #print("%02X "%struct.unpack("B", arr[i:i+1]))
                    #r = recordLength+dataLength-i-1
                    #value += "%02X "%struct.unpack("B", arr[i:i+1])[0]
                data = arr[recordLength:recordLength+dataLength]
                # Save the ABB version
                if abb_version_object:
                    print("abb_version_object True in 0x0D")
                    print("data", data)
                    abb_version_data = data
                value = "".join("%02X "%(i) for i in data[::-1])   # Byte ouput.
                    #value = "".join(chr(i) for i in data[::-1])         # Char ouput.
            elif LVAR >= 0xC0 and LVAR <= 0xCF: # Positive BCD number.
                dataLength = LVAR & 0x0F
                #dataTypeStr="Positive BCD number"
                value=""
                for i in range(dataLength):
                    value += "%02X"%int(arr[recordLength+dataLength-i-1],16)
            elif LVAR >= 0xD0 and LVAR <= 0xDF: # Negative BCD number.
                dataLength = LVAR & 0x0F
                #dataTypeStr="Negative BCD number"
                value="-"
                for i in range(dataLength):
                    value += "%02x"%int(arr[recordLength+dataLength-i-1],16)
                value = value.hex().upper()
            elif LVAR >= 0xE0 and LVAR <= 0xEF:  # Binary number with (LVAR - 0xE0) bytes.
                dataLength = LVAR & 0x0F
                #dataTypeStr="Binary number, showing as bytes"
                value=""
                for i in range(dataLength):
                    value += "%02x"%int(arr[recordLength+i],16)
                value = value.upper()
            elif LVAR >= 0xF0 and LVAR <= 0xF4: # Binary number with 4*(LVAR - 0xEC) bytes.
                dataLength = LVAR & 0x0F
                #dataTypeStr="Binary number, showing as 4 bytes"
                value=""
                for i in range(dataLength):
                    value += "%02x"%int(arr[recordLength+i],16)
                value = value.upper()
            elif LVAR == 0xF5: # Binary number with 48 bytes.
                dataLength = 48
                #dataTypeStr="Binary number with 48 bytes"
                value = "Binary 48 byte value"
            elif LVAR == 0xF6: # Binary number with 64 bytes.
                #dataLenght = 64
                #dataTypeStr="Binary number with 64 bytes"
                value = "Binary 68 byte value"
            else: # Reserved
                #dataLenght = 0
                #dataTypeStr="Reserved"
                value = "Reserved"
            recordLength += dataLength
        elif dataType==0x0E: # BCD12.
            dataTypeStr="BCD12"
            value=bcdtoval(arr[recordLength:recordLength+6])
            recordLength += 6

        # Decode value to date or time.
        if isDateTime and (dataType == 0x02 or dataType == 0x03 or dataType == 0x04 or dataType == 0x06 or dataType == 0x0E):
            value, vib = DecodeDateTime(dataType, value)
    except:
        value = "Error"
        dataTypeStr = "Error"

    return_map = {
        "value": value,
        "dataTypeStr": dataTypeStr,
        "recordLength": recordLength,
        "function": function,
        "mfct_data": mfct_data,
        "more_telegrams": more_telegrams,
        "vib": vib,
        "exponent": exponent,
        "templateVIB": templateVIB,
        "abb_version_data": abb_version_data
    }
    return return_map

# Take input and select the records that should be saved to a template.
def SelectDataRecordNumbers():
    print("Type in the data record numbers, example '1 2 5 10': ", end="")
    inp = input()
    if inp == "":    # Exit if no records is given.
        return
    numbers = inp.split()   # ToDo: Return only destinct values, remove duplicates.
    # ToDo:
    # Check if all numbers are valid integers.
    # Return list of integers.
    return numbers

def SelectDataRecordsByGivenNumbers(numbers, records):
    selected = {}    # Storage for all selected data records.
    # Gets the data records according to selected inputs.
    for number in numbers:
        number = int(number)
        if number in records:
            selected[number] = records[number]
    return selected

# Select data records for a template.
#def SelectDataRecords(records):
#    numbers = SelectDataRecordNumbers()
#    selected = SelectDataRecordsByGivenNumbers(numbers, records)
#    #recordsForTemplate.update(selected)

# Save a telegram's record in a JSON-file.
def SaveTelegramInJSonFile(records, telegramCounter, secondaryAddress,  statusByte,  statusText,  configWord, more_telegrams):
    #templateFolderPath = "/tmp/"
    #SetupTemplateFolder(templateFolderPath)
    templateFilePath = "/tmp/browseOutput.json" #= CreateTemplateFile(templateFolderPath, secondaryAddress[1], secondaryAddress[2], secondaryAddress[3])
    json = StoreRecordsAsJSON(records, telegramCounter, secondaryAddress,  statusByte,  statusText,  configWord)
    CreateJSONTemplate(json, templateFilePath)
    if more_telegrams:
        print("Telegram number", telegramCounter)
    else:
        if telegramCounter == 1: # English grammatic, telegram in plural or not
            print("JSON-file with the data records for %s telegram stored at: %s"%(telegramCounter, templateFilePath))
        else:
            print("JSON-file with the data records for %s telegrams stored at: %s"%(telegramCounter, templateFilePath))
        print("")

def SaveTelegramInCsvFile(records, telegramCounter, secondaryAddress, templateParams,  statusByte, statusText, more_telegrams, abb_version_string):
    templateFileName = "/tmp/templates_tmp.csv"
    data = StoreRecordsAsCSV(records, telegramCounter, secondaryAddress, templateParams,  statusByte, statusText, more_telegrams, abb_version_string)
    CreateCSVTemplate(data, templateFileName)

# Creates the template data from selected records.
def StoreRecordsAsJSON(records, telegramCounter, secondaryAddress,  statusByte,  statusText,  configWord):
    #print("json reocrds", records)
    #json.dumps({'device':{'name':'4129.03.31','telegram':1,'address':'16809984.4129.03.31','tag':[{'name':'Fabrication','record':'1','datatype':'BCD8'},{'name':'Volts V E-1','record':'2','datatype':'INT16'},{'name':'Current A E-4','record':'3','datatype':'INT16'}]}})
    #{"device": {"name": "4129.03.31", "telegram": 1, "address": "16809984.4129.03.31", "tag": [{"name": "Fabrication", "record": "1", "datatype": "BCD8"}, {"name": "Volts V E-1", "record": "2", "datatype": "INT16"}, {"name": "Current A E-4", "record": "3", "datatype": "INT16"}]}}
    deviceName = "%s_%s_%s"%(MfctToStr(int(secondaryAddress[1], 16)), secondaryAddress[2], secondaryAddress[3])
    secondaryAddress = "%s.%s.%s.%s"%(secondaryAddress[0], secondaryAddress[1], secondaryAddress[2], secondaryAddress[3])
    # ToDo: Replace all ' with " in the output JSON-file.
    # Returns unix timestamp for current time
    #timestamp = time.time()
    # Get datetime object in local time
    #now = datetime.datetime.fromtimestamp(timestamp)

    named_tuple = time.localtime() # get struct_time
    now = time.strftime("%Y-%m-%d %H:%M:%S", named_tuple)

    # Return unix timestamp from datetime object
    #timestamp = d.timestamp()
    jsonData = ""
    jsonData += "{"
    jsonData += "'device':"
    jsonData += "{"
    jsonData += "'name':'%s','telegram':%s,'adress':'%s','prim_adr':%s,'name2':'%s', 'type':'%s', 'time':'%s', 'statusByte':'%s', 'statusText':'%s', 'configWord':'%s'},"%(deviceName, telegramCounter, secondaryAddress, primaryAddress, nameGlobal,  typeGlobal,  now,  statusByte,  statusText,  configWord) # La till en måsvinge } efter adress
    jsonData += "'tag':"
    jsonData += "["

    length = len(records) - 1
    templateVIBs=[] # Check if the same tag has appeared earlier
    for count,record in enumerate(records): #Ändrade så att sista kommatecknet försvinner i json-outputen
        expOutput = ""
        tagName = records[record][6].replace("\'", "\\'")
        dataType = records[record][1]
        value = records[record][0]
        vif = records[record][8]
        function = records[record][3]
        i = countItems(vif, templateVIBs)
        templateVIBs.append(vif)
        if i>0:
            vif += '%d'%i
        if record <= length:
            expOutput = records[record][7]

        jsonData += "{"
        jsonData += "'name':'%s','value':'%s','exponent':'%s','record':'%s','datatype':'%s', 'vif':'%s', 'function':'%s'"%(tagName,value,expOutput,record,dataType,vif,function)
    
        if count < length:
            jsonData += "},"
        else:
            jsonData += "}"

    jsonData += "]"
    jsonData += "}"
    jsonData = jsonData.replace('\'',  '\"')

    return jsonData

def countItems(val,  list):
    i = 0
    for item in list:
        if val == item:
            i += 1
    return i

# Get the datatype for modbus readout
def getDataType(dataType):
    if dataType == "BCD8" or dataType == "INT24" or dataType == "INT32" or dataType == "BCD6":
        return 3
    elif dataType == "INT48" or dataType == "INT64" or dataType == "BCD12":
        return 20
    elif dataType == "REAL32":
        return 4
    elif dataType == "REAL64":
        return 5
    elif dataType == "BCD4" or dataType == "INT16" or dataType == "INT8":
        return 2
    return ""

# Save a telegram's record in a CSV-file.
def StoreRecordsAsCSV(records, telegramNumber, secondaryAddress, templateParams, statusByte, statusText, more_telegrams, abb_version_string):
    print("records", records)
    st = MfctToStr(int(secondaryAddress[1], 16))
    saddr = "%s.%s.%s.%s"%(secondaryAddress[0], secondaryAddress[1], secondaryAddress[2], secondaryAddress[3]) #Add fab no XXXXXXXXXXXXXXXXXXXXXXXX
    # ToDo: Replace all ' with " in the output JSON-file.
    # Returns unix timestamp for current time
    # Get datetime object in local time
    csvData = []

    # Create a mfct_spec_ver tuple to compare to the MANUFACTURER_SPECIFIC_VERSION_SET in weblib.py
    mfct_spec_ver = (secondaryAddress[1], secondaryAddress[2], secondaryAddress[3])

    device = ['']*45
    device[0]="[Device]"
    device[1] = 'channel' # templateParams['channel_name']
    device[3] = "%s_%s_%s"%(st, secondaryAddress[2], secondaryAddress[3]) # Template device Name
    if mfct_spec_ver in MANUFACTURER_SPECIFIC_VERSION_SET:
        device[3] = "_".join([device[3], abb_version_string])
    device[13] = 3 # Number of telegrams
    device[19] = saddr
    if (templateParams['format'] == "1" or templateParams['format'] == "2" or templateParams['format'] == "4") and templateParams['tag_type'] == "8":
        device[21] = "f"
    device[24] = int(templateParams['time_stamp']) + int(templateParams['time_stamp_telegram'])
    r =  templateParams['read_period']
    if r.isnumeric():
        device[41] = 60*int( templateParams['read_period'])
    r =  templateParams['read_offset']
    if r.isnumeric():
        device[42] = 60*int( templateParams['read_offset'])
    csvData.append(device)

    #Appending the statusbyte to the template file
    tag = ['']*45
    tag[0] = "[Tag]"
    tag[1] = 'channel' # templateParams['channel_name']
    tag[3] = "%s_%s_%s"%(st, secondaryAddress[2], secondaryAddress[3]) # Template device Name
    if mfct_spec_ver in MANUFACTURER_SPECIFIC_VERSION_SET:
        tag[3] = "_".join([tag[3], abb_version_string])
    tag[4] = "StatusByte"
    #tag[19] = record
    tag[20] = 22
    tag[21] = 8 # String
    tag[22] = 1 # Readable
    tag[25] = 0
    tag[36] = 0
    csvData.append(tag)

    #Appending the Modbus statusbyte to the template file
    tag = ['']*45
    tag[0] = "[Tag]"
    tag[1] = 'channel' # templateParams['channel_name']
    tag[3] = "%s_%s_%s"%(st, secondaryAddress[2], secondaryAddress[3]) # Template device Name
    if mfct_spec_ver in MANUFACTURER_SPECIFIC_VERSION_SET:
        tag[3] = "_".join([tag[3], abb_version_string])
    tag[4] = "StatusByte_Mod"
    #tag[19] = record
    tag[20] = 22
    tag[21] = 2 # INT16, Smallest value in modbus register
    tag[22] = 1 # Readable
    tag[25] = 0
    tag[36] = 0
    csvData.append(tag)

    templateVIBs=[] # Check if the same tag has appeared earlier

    for record in records:
        if records[record][4] == True or records[record][0] == "Error": #If manufacturer specific or error, do nothing
            continue
        templateVIB = records[record][8]
        dataType = records[record][1]
        i = countItems(templateVIB, templateVIBs)
        templateVIBs.append(templateVIB)
        if i>0:
            templateVIB += '%d'%i
        tag = ['']*45
        tag[0] = "[Tag]"
        tag[1] = 'channel'
        tag[3] = "%s_%s_%s"%(st, secondaryAddress[2], secondaryAddress[3]) # Template device Name
        if mfct_spec_ver in MANUFACTURER_SPECIFIC_VERSION_SET:
            tag[3] = "_".join([tag[3], abb_version_string])
        tag[4] = templateVIB
        tag[19] = record
        tag[20] = templateParams['tag_type']
        tag[21] = 8 # String
        tag[22] = 1 # Readable

        csvData.append(tag)

        #Modbus tag
        #print("dataType {}".format(dataType))
        tag = ['']*45
        tag[0] = "[Tag]"
        tag[1] = 'channel'
        tag[3] = "%s_%s_%s"%(st, secondaryAddress[2], secondaryAddress[3]) # Template device Name
        if mfct_spec_ver in MANUFACTURER_SPECIFIC_VERSION_SET:
            tag[3] = "_".join([tag[3], abb_version_string])
        tag[4] = templateVIB + "_Mod"
        tag[19] = record
        #tag[20] = templateParams['tag_type']
        tag[20] = 1
        tag[21] = getDataType(dataType)
        
        tag[22] = 1 # Readable
        tag[36] = 0 # 
        csvData.append(tag)

    if int(templateParams['time_stamp_telegram']) == 2:
        tag = ['']*45
        tag[0] = "[Tag]"
        tag[1] = 'channel'
        tag[3] = "%s_%s_%s"%(st, secondaryAddress[2], secondaryAddress[3]) # Template device Name
        tag[4] ="time"
        tag[19] = int(templateParams['time_record_no'])
        tag[20] = 1 # Tagtype
        tag[21] = 8 # String
        tag[22] = 1 # Readable
        tag[24] = int(templateParams['time_stamp']) + int(templateParams['time_stamp_telegram'])
        csvData.append(tag)

    return csvData

# Create the csv template.
def CreateTemplate(data, templateFilePath):
    #file = open(templateFilePath, 'a+')
    with open(templateFilePath,  'w') as fil:
        for line in data:
            fil.write(line)

# Create the json template.
def CreateJSONTemplate(data, templateFilePath):
    #file = open(templateFilePath, 'a+')
    with open(templateFilePath,  'w') as fil:
        for line in data:
            fil.write(line)

# Create the csv file
def CreateCSVTemplate(data, templateFilePath):
    with open(templateFilePath,'w', newline='') as cf:
        writer = csv.writer(cf, delimiter=',', quotechar='\"', quoting=csv.QUOTE_MINIMAL)
        for row in data:
            writer.writerow(row)

def CreateSlvSelect(idnum, mfct=0xFFFF, ver=0xFF, med=0xFF, fab = None):
    req1 = b"\x53\xFD\x52"
    req1 += struct.pack("<IHBB",idnum, mfct, ver, med)
    if fab:
        req1 += b"\x0C\x78" + struct.pack("<I", fab)
    req = b"\x68" + struct.pack("BB",len(req1),len(req1)) + b"\x68" + req1
    req += struct.pack('B',CalcBcc(req[4:]))+b"\x16"
    return req

def CreateShortFrame(cfield, afield):
    req   = b"\x10"+struct.pack("BB",cfield, afield)
    req += struct.pack("B",CalcBcc(req[1:])) + b"\x16"
    return req

def VerifyE5(resp): # Check for none ACK response.
    retval = ""
    if len(resp) == 0:
        retval = "No SND_NKE response"
        print(retval)
        return False
    elif len(resp) != 1 or resp[0] != 0xE5:
        retval = "Invalid SND_NKE_RESPONSE, %02X"%resp[0]
        print(retval)
        return False
    return True

def CreateAppReset(addr,  subcode=None):
    req = b"\x53" + struct.pack("B",addr) + b"\x50"
    if subcode:
        req += struct.pack("B", subcode)
    req = b"\x68" + struct.pack("BB", len(req), len(req)) + b"\x68" + req;
    req += struct.pack("B", CalcBcc(req[4:]) ) + b"\x16"
    return req

def VerifyLongFrame(frame):
    generalApplicationError = False
    if len(frame) == 0:
        print("No response received")
        return generalApplicationError, False
    if (frame[0] != 0x68 or frame[3] != 0x68): # Check frame characters.
        print("Incorrect frame characters")
        return generalApplicationError, False
    if (frame[1] != frame[2] or (frame[1] != len(frame)-6 and frame[1] != len(frame)-262)): # Check length. Allow for too long telegrams GG 2021-12-21
        print("Incorrect message length")
        return generalApplicationError, False
    if (CalcBcc(frame[4:len(frame)-2]) != frame[len(frame)-2]): # Check checksum.
        print("Incorrect checksum")
        return generalApplicationError, False
    if len(frame) < 21 and frame[6] == 0x70:
        generalApplicationError = True
        return generalApplicationError, True
    if len(frame) < 21:  # Check number of received bytes.
        print("Too few bytes received")
        print(len(frame))
        return generalApplicationError, False
    return generalApplicationError, True

def getIniVal(inif,  group,  item,  default = ''):
    retval = ''
    try:
        retval = inif.get(group, item)
    except:
        pass
    if not retval:
        retval = default
    #print("retval = %s"%retval)
    return retval

def browse(ip, port, init, adr, numberOfTelegrams=1, slv_sel_snd_nke=False, deselect=False):
    try:
        #The program must work without a Browse.ini file XXXXXXXXXXXXXXXXXX
        browseini = configparser.RawConfigParser()
        browseini.read("/tmp/Browse.ini")
        del exponentGlobal[:]
        templateParams = {}
        templateParams['format'] = getIniVal(browseini, "BROWSE_PARAMETERS",  "format", "FORMAT3")
        templateParams['channel_name'] = getIniVal(browseini, "BROWSE_PARAMETERS",  "channel_name", "pi900")
        templateParams['time_stamp'] = getIniVal(browseini, "BROWSE_PARAMETERS",  "time_stamp", "24")
        templateParams['read_period'] = getIniVal(browseini, "BROWSE_PARAMETERS",  "read_period")
        templateParams['read_offset'] = getIniVal(browseini, "BROWSE_PARAMETERS",  "read_offset")
        templateParams['tag_type'] = getIniVal(browseini, "BROWSE_PARAMETERS",  "tag_type", "8")
        templateParams['time_stamp_telegram'] = getIniVal(browseini, "BROWSE_PARAMETERS",  "time_stamp_telegram", '0')
        templateParams['time_record_no'] = getIniVal(browseini, "BROWSE_PARAMETERS",  "time_record_no", '4')

        mbushubini_def = configparser.RawConfigParser()
        mbushubini_def.read("/config/mbushub_default.ini")
        mbushubini = configparser.RawConfigParser()
        mbushubini.read("/config/mbushub.ini")

        parser = Configuration(mbushubini_def, mbushubini)
        sercomport = parser.get(float, "MasterPort", "timeout")

        time_out = (sercomport/1000) + 0.2
        print("sercomport", time_out)

        s = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
        s.settimeout(time_out)
        response=""

        # Use either primary or secondary address.
        if len(adr) < 4: # Primary address. Well, needs to be documented. 123 is a valid secondary address 00000123.FFFF.FF.FF
            adr = int(adr)
        else:   # Secondary address
            # Request format: 68 L L 68 C ADR CI XX XX XX XX MM MM VV MM CS 16.
            secaddr = adr.split(':')
            idnum, mfct, ver, med = secaddr[0].split('.')
            idnum = int(idnum,16)
            mfct  = int(mfct,16)
            ver   = int(ver,16)
            med   = int(med,16)
            fabr  = None
            # Enhanced selection in SLV_SEL request?
            if (len(secaddr) > 1):
                fabr = int(secaddr[1],16)

            if slv_sel_snd_nke == True:
                print("slv_sel_snd_nke")
                request = CreateSlvSelect(idnum, mfct, ver, med, fabr)
                s.sendto(request, (ip, port))
                response = s.recv(1024)

                if not VerifyE5(response):
                    return_map = {
                        'success': False,
                        'status-string': "No SLV_SEL resp"
                    }
                    return return_map

                snd_nke_req = CreateShortFrame(0x40, 253)
                s.sendto(snd_nke_req, (ip, port))
                snd_nke_resp = s.recv(1024)

                if not VerifyE5(response):
                    return_map = {
                        'success': False,
                        'status-string': "No SNK_NKE resp"
                    }
                    return return_map

                request = CreateSlvSelect(idnum, mfct, ver, med, fabr)
                s.sendto(request, (ip, port))
                response = s.recv(1024)

                if not VerifyE5(response):
                    return_map = {
                        'success': False,
                        'status-string': "No SLV_SEL resp"
                    }
                    return return_map
            else:
                print("slv_sel_only")
                request = CreateSlvSelect(idnum, mfct, ver, med, fabr)
                s.sendto(request, (ip, port))
                response = s.recv(1024)

                if not VerifyE5(response):
                    return_map = {
                        'success': False,
                        'status-string': "No SLV_SEL resp"
                    }
                    return return_map
            adr = 253

        # Send SND_NKE and receive ACK.
        if init == "s":
            request = CreateShortFrame(0x40,  adr)
            s.sendto(request, (ip, port))
            response = s.recv(1024)
            if not VerifyE5(response):
                return_map = {
                    'success': False,
                    'status-string': "No SNK_NKE resp"
                }
                return return_map
        # Send APP_RST and receive ACK.
        if init == "a":
            request = CreateAppReset(adr)
            s.sendto(request, (ip, port))
            response = s.recv(1024)
            if not VerifyE5(response):
                return_map = {
                    'success': False,
                    'status-string': "No APP_RES resp"
                }
                return return_map

        toggle = 0x7B           # Toggle field for REQ_UD2.
        dataRecord = 0          # Init data record index.
        telegramCounter = 1     # Telegram counter.
        secondaryAddress = None   # Secondary address.
        run = True              # Keep on reading the M-Bus meter.
        records = {}            # List of all records for a telegram.
        more_telegrams = True
        statusByte = ""
        statusText = ""
        configWord = ""

        # The secondary address variables from the RSP_UD
        # These will be returned so the input secondary-address can
        # be compared with the output-secondary-address
        identification = ""
        manufacturer = ""
        version = ""
        medium = ""
        name = ""
        m_type = ""
        abb_version_raw = ""
        abb_version_string = ""
        primaryAddress = ""

        while telegramCounter <= int(numberOfTelegrams) and more_telegrams: # Send REQ_UD2 and receive RSP_UD.
            try:
                request = CreateShortFrame(toggle,  adr)
                s.sendto(request, (ip, port))
                response = s.recv(1024)
            except Exception as e:
                print("Failed to read the M-Bus meter", e)

                return_map = {
                    'success': False,
                    'status-string': "No REQ_UD resp"
                }

                return return_map

            toggle = 0x5B if toggle == 0x7B else 0x7B
            responseLength=len(response)
            generalApplicationError, aCorrectResponse = VerifyLongFrame(response)

            if not aCorrectResponse:
                break
            elif generalApplicationError and aCorrectResponse:
                primaryAddress = response[5]
                secondaryAddress = identification, manufacturer, version, medium, name, m_type = SecondaryAddress(response[7:15])
                statusByte = ""
                statusText = "CI-field 0x70, General Application Error"
                configWord = ""
            elif not generalApplicationError and aCorrectResponse:
                primaryAddress = response[5]
                identification, manufacturer, version, medium, name, m_type = SecondaryAddress(response[7:15])
                secondaryAddress = (identification, manufacturer, version, medium, name, m_type)
                statusByte = "%02X"%response[16]
                statusText = Status(response[16])

                configWord = response[17:19]
                configWord = bytearray(configWord)
                configWord.reverse()
                configWord = "".join("%02X" % b for b in configWord)

            OPCname = MfctToStr(int(manufacturer, 16))
            if OPCname != "KAM":
                OPCname += version + medium # "ABB2002" example
            print("\n --- Telegram %s ---\n"%(telegramCounter))
            print("\tPrimary address:\t%s"%(primaryAddress))
            print("\tSecondary address:\t%s.%s.%s.%s [%s] [%s]"%(identification, manufacturer, version, medium, name, m_type))
            print("\tStatus:\t\t\t%s %s"%(statusByte, statusText))
            print("")
            print("{:3s} {:7s} {:40s} {:20s}".format("#", "Type", "Value", "Physical quantity"))
            print("-------------------------------------------------------------------------------")

            indexInTelegram = 19

            # ============================================================================
            # === Iterate over all objects in the telegram and display them in a table ===
            # ============================================================================
            mfct_data = False

            print("responseLength", responseLength)
            if not generalApplicationError:
                while indexInTelegram + 2 < responseLength:

                    returnMap = decode_field(response[indexInTelegram:], OPCname, telegramCounter)  # Decode the object/record.
                    value = returnMap["value"]
                    dataTypeStr = returnMap["dataTypeStr"]
                    recordLength = returnMap["recordLength"]
                    function = returnMap["function"]
                    mfct_data = returnMap["mfct_data"]
                    more_telegrams = returnMap["more_telegrams"]
                    vib = returnMap["vib"]
                    exponent = returnMap["exponent"]
                    templateVIB = returnMap["templateVIB"]
                    abb_version_raw = returnMap["abb_version_data"]

                    if len(abb_version_raw) > 0 and telegramCounter == 1:
                        abb_version_string = abb_version_raw.decode("utf-8")[::-1]
                        print("abb_version_string", abb_version_string)

                    indexInTelegram +=  recordLength   # Point to next object in telegram.

                    if value == None:
                        continue
                    #if templateVIB == 'mfct':
                    #    continue
                    dataRecord += 1
                    records[dataRecord] = (value, dataTypeStr, recordLength, function, mfct_data, more_telegrams, vib, exponent, templateVIB, dataRecord)
                    print("{:3s} {:7s} {:40s} {:20s}".format(str(dataRecord), dataTypeStr, str(value)+exponent, vib))
                    exponentGlobal.append(exponent)

            print("")

            if (more_telegrams):
                telegramCounter += 1
            else:
                break

        print("")
        SaveTelegramInJSonFile(records, telegramCounter, secondaryAddress,  statusByte,  statusText, configWord, more_telegrams)
        SaveTelegramInCsvFile(records, telegramCounter, secondaryAddress, templateParams,  statusByte, statusText, more_telegrams, abb_version_string)

        # Addition - Always make a SND_NKE to address 253 after REQ_UD2
        # Some wired meters do not understand SLV_SEL with fabrication-number
        # therefore a SND_NKE is needed to deselect the meter

        if deselect == True:
            print("deselect")
            snd_nke_req = CreateShortFrame(0x40, 253)
            s.sendto(snd_nke_req, (ip, port))
            snd_nke_resp = s.recv(1024)

        print(identification, manufacturer, version, medium, primaryAddress, name, m_type, abb_version_string)
        return_map = {
            'identification': identification,
            'manufacturer': manufacturer,
            'version': version,
            'medium': medium,
            'primaryAddress': primaryAddress,
            'name': name,
            'm_type': m_type,
            'abb_version_string': abb_version_string
        }
        return return_map

    except Exception as e:
        print("error in Browse",str(e))
        traceback.print_exc()

if __name__ == "__main__":
    if len(sys.argv) < 5:
        print("Usage %s <IP-address> <IP-Port> <init> <M-Bus address>"%sys.argv[0])
        sys.exit(0)

    ip = sys.argv[1]
    port = int(sys.argv[2])
    init = sys.argv[3]
    adr = sys.argv[4]

    if len(sys.argv) > 5:
        numberOfTelegrams = int(sys.argv[5])
        select = int(sys.argv[6])
        deselect = int(sys.argv[7])
        browse(ip, port, init, adr, numberOfTelegrams, select, deselect)
    else:
        browse(ip, port, init, adr)
