#!flask/bin/python3
from flask import Flask, jsonify, abort, request, make_response, url_for,  Blueprint
from flask_httpauth import HTTPBasicAuth
import os
import time
import configparser
from werkzeug.utils import secure_filename

webapi_blueprint = Blueprint('webapi', __name__, url_prefix='/api/v1.0')

auth = HTTPBasicAuth()

@auth.get_password
def get_password(username):
    if username == 'Admin':
        return 'Admin'
    return None

@auth.error_handler
def unauthorized():
    return make_response(jsonify( { 'error': 'Unauthorized access' } ), 403)
    # return 403 instead of 401 to prevent browsers from displaying the default auth dialog

@webapi_blueprint.errorhandler(400)
def not_found(error):
    return make_response(jsonify( { 'error': 'Bad request' } ), 400)

def setPi900Time(timestr):
    try:
        dt = time.strptime(timestr, '%Y-%m-%dT%H:%M:%S')
        os.system(time.strftime("date --set=\"%Y-%m-%d %H:%M:%S\"\n",dt))
        os.system("hwclock -u -w --rtc /dev/rtc0\n")
        os.system("hwclock -u -w --rtc /dev/rtc1\n")
    except:
        return False
    return True

def deletePi900Ntp():
    os.system("rm -f /etc/ntp.conf")
    return True

def savePi900Ntp(server):
    with open("/etc/ntp.conf",'w') as f:
        f.write("""server %s iburst
server 2.pool.ntp.org iburst
server 3.pool.ntp.org iburst
# Allow only time queries, at a limited rate, sending KoD when in excess.
# Allow all local queries (IPv4, IPv6)
restrict default nomodify nopeer noquery limited kod
restrict 127.0.0.1
restrict [::1]\n"""%server)
    os.system("sync;/etc/init.d/S49ntp restart &")
    return True

# curl -uAdmin:Admin -X POST localhost:5000/api/v1.0/license -H "Content-Type: application/json" -H "Accept: application/json" -d '{"licensestring":"EDE32C2AB8DF6E05A014879DD98FB18C327B5AED64F10F52CBCE7B346BA5D5E4000EA71CA648187CB6D27221C87CB997DC0A37DE2E42025E853DF7D9EC39C1AE"}'
# curl -uAdmin:Admin -X POST localhost:5000/api/v1.0/license
# set license works.
@webapi_blueprint.route("/license", methods=['POST'])
def license():
    if request.method == 'POST':
        if request.is_json:
            req = request.get_json()
            licstr = req['licensestring']
            ret = os.system("/binary/mbushub.elf -c %s"%licstr)
            if ret:
                return make_response(jsonify( { 'status': 'error', 'message': 'Incorrect license string' } ), 500)
            with open('/config/mbushub_lic.ini','w') as cf:
                cf.write("[Main]\nlicstr=%s\n"%licstr)
                cf.close()
            os.system("/bin/sync")
            os.system("killall mbushub.elf")
            os.system("/binary/mbushub.elf >/dev/null 2>&1 &")
            ret = os.system("/binary/mbushub.elf -c %s"%licstr)
            if ret:
                return make_response(jsonify( { 'status': 'error', 'message': 'Saving license failed' } ), 500)
            return jsonify({'status': 'OK'})
        else:
            conf = configparser.RawConfigParser()
            licstr = None
            try:
                conf.read("/config/mbushub_lic.ini")
                licstr = conf.get("Main","licstr")
                return jsonify({'status': 'OK','licensestring': licstr})
            except:
                pass
    abort(400)

#curl -uAdmin:Admin -X POST localhost:5000/api/v1.0/ntp -H "Content-Type: application/json" -H "Accept: application/json" -d '{"server":"delete"}'
#curl -uAdmin:Admin -X POST localhost:5000/api/v1.0/ntp -H "Content-Type: application/json" -H "Accept: application/json" -d '{"server":"0.pool.ntp.org"}'
@webapi_blueprint.route("/ntp", methods=['POST'])
def ntp():
    if request.method == 'POST' and request.is_json:
        req = request.get_json()
        if req['server'] == 'delete':
            deletePi900Ntp()
            return jsonify({'status': 'OK'})
        else:
            savePi900Ntp(req['server'])
            return jsonify({'status': 'OK', 'server': req['server']})
    abort(400)

# curl -uAdmin:Admin -X POST localhost:5000/api/v1.0/datetime -H "Content-Type: application/json" -H "Accept: application/json" -d '{"utctime":"2019-05-06T12:15:15"}'
#{ "status": "ok",  "utctime": "2019-10-10T18:37:14" }
# curl -uAdmin:Admin -X POST localhost:5000/api/v1.0/datetime
#{ "status": "ok",  "utctime": "2019-10-10T18:37:54"}
@webapi_blueprint.route("/datetime", methods=['POST'])
@auth.login_required
def datetime():
    if request.is_json:
        req = request.get_json()
        if not setPi900Time(req['utctime']):
            abort(400)
    t = time.strftime("%Y-%m-%dT%H:%M:%S",time.gmtime())
    return jsonify({'status': 'ok', 'utctime': t})

# curl -uAdmin:Admin -X POST localhost:5000/api/v1.0/reboot
@webapi_blueprint.route("/reboot", methods=['POST'])
@auth.login_required
def reboot():
    if request.method == 'POST':
        os.system("/sbin/reboot")
        return jsonify({'status': 'OK'})
    abort(400)

# curl -uAdmin:Admin -i -X POST -H "Content-Type: multipart/form-data" -F "file=@test.csv" localhost:5000/api/v1.0/uploadconfig
@webapi_blueprint.route("/uploadconfig", methods=['POST'])
@auth.login_required
def uploadConfigfile():
    file = request.files['file']
    if file:
        filename = secure_filename(file.filename)
        file.save(os.path.join("/config", filename))
        return jsonify({'status': 'OK'})
    abort(400)


