from shutil import copyfile
from uuid import getnode
import configparser
import os, re, time, glob
from werkzeug.utils import secure_filename
from flask import Flask, render_template, request, send_file, redirect, url_for, render_template_string, after_this_request, json, make_response,jsonify
from weblib import *
import codecs
import subprocess
import traceback

app = Flask(__name__, static_url_path='')
app.jinja_env.add_extension('jinja2.ext.loopcontrols')
app.jinja_env.add_extension('jinja2.ext.do')

mbushubInstalledCorrectly = True
quickpostInstalledCorrectly = True
wirelessInstalledCorrectly = True
mbusAsciiMQTTInstalledCorrectly = True
modBMbusInstalledCorrectly = True
hanInstalledCorrectly = True
alarmServerInstalledCorrectly = True
dlms2mbusInstalledCorrectly = True

if os.path.exists("mbushub/mbushub.py"):
    try:
        from mbushub.mbushub import mbushub_blueprint
        app.register_blueprint(mbushub_blueprint)
    except Exception as e:
        traceback.print_exc()
        mbushubInstalledCorrectly = False

if os.path.exists("wireless/wireless.py"):
    try:
        from wireless.wireless import wireless_blueprint
        app.register_blueprint(wireless_blueprint)
    except:
        wirelessInstalledCorrectly = False
        traceback.print_exc()

if os.path.exists("quickpost/quickpost.py"):
    try:
        from quickpost.quickpost import quickpost_blueprint
        app.register_blueprint(quickpost_blueprint)

    except:
        quickpostInstalledCorrectly = False
        traceback.print_exc()

if os.path.exists("mbusascii2mqtt/mbusascii2mqtt.py"):
    try:
        from mbusascii2mqtt.mbusascii2mqtt import mbusascii2mqtt_blueprint
        app.register_blueprint(mbusascii2mqtt_blueprint)

    except:
        mbusAsciiMQTTInstalledCorrectly = False
        traceback.print_exc()

if os.path.exists("modbus2mbus/modbus2mbus.py"):
    try:
        from modbus2mbus.modbus2mbus import modbus2mbus_blueprint
        app.register_blueprint(modbus2mbus_blueprint)

    except:
        modBMbusInstalledCorrectly = False
        traceback.print_exc()

if os.path.exists("han2mbus/han2mbus.py"):
    try:
        from han2mbus.han2mbus import han2mbus_blueprint
        app.register_blueprint(han2mbus_blueprint)

    except:
        hanInstalledCorrectly = False
        traceback.print_exc()

if os.path.exists("alarmserver/alarmserver.py"):
    try:
        from alarmserver.alarmserver import alarmserver_blueprint
        app.register_blueprint(alarmserver_blueprint)

    except:
        alarmServerInstalledCorrectly = False
        traceback.print_exc()

if os.path.exists("dlms2mbus/dlms2mbus.py"):
    try:
        from dlms2mbus.dlms2mbus import dlms2mbus_blueprint
        app.register_blueprint(dlms2mbus_blueprint)
    except:
        dlms2mbusInstalledCorrectly = False
        traceback.print_exc()

try:
    from core.search import search_blueprint
    app.register_blueprint(search_blueprint)
except:
    raise

def allowed_file(filename):
    return '.' in filename and \
           filename.rsplit('.', 1)[1].lower() in ['csv','ini','xml']

@app.route("/incorrect_installation",  methods=['GET'])
def incorrectInstallation():
    software = request.args.get("software")
    return render_template("problems_install_software.html",  software=software)

@app.route("/")
def index():
    identification = ""
    try:
        with open("/var/www1/identification.py") as file:
            identification=file.read()
            identification = identification.strip()
    except Exception as e:
        print("No identification set. "+str(e))

    info = {
        'pi900_type' : pi900_type,
        'identification' : identification
    }
    return render_template('index.html',  **info)

@app.route("/header")
def header():
    identification = ""
    try:
        with open("/var/www1/identification.py") as file:
            identification=file.read()
            identification = identification.strip()
    except Exception as e:
        print("No identification set. "+str(e))

    info = {
        'pi900_type' : pi900_type,
        'identification' : identification
    }
    return render_template('header.html',  **info)

@app.route("/menu.html")
def menu():
    info = {
        'mbushubInstalled' : os.path.exists("mbushub/mbushub.py"),
        'quickpostInstalled' : os.path.exists("quickpost/quickpost.py"),
        'wirelessInstalled' : os.path.exists("wireless/wireless.py"),
        'mbusascii2mqttInstalled' : os.path.exists("mbusascii2mqtt/mbusascii2mqtt.py"),
        'modbus2mbusInstalled' : os.path.exists("modbus2mbus/modbus2mbus.py"),
        'han2mbusInstalled' : os.path.exists("han2mbus/han2mbus.py"),
        'alarmserverInstalled' : os.path.exists("alarmserver/alarmserver.py"),
        'mbushubInstalledCorrectly' : mbushubInstalledCorrectly,
        'quickpostInstalledCorrectly' : quickpostInstalledCorrectly,
        'wirelessInstalledCorrectly' : wirelessInstalledCorrectly,
        'mbusAsciiMQTTInstalledCorrectly' : mbusAsciiMQTTInstalledCorrectly,
        'modBMbusInstalledCorrectly' : modBMbusInstalledCorrectly,
        'hanInstalledCorrectly' : hanInstalledCorrectly,
        'alarmServerInstalledCorrectly' : alarmServerInstalledCorrectly,
        'dlms2mbusInstalled' : os.path.exists("dlms2mbus/dlms2mbus.py"),
        'dlms2mbusInstalledCorrectly' : dlms2mbusInstalledCorrectly
    }
    return render_template('menu.html',  **info)

@app.route("/start_page")
def startPage():

    #Administration card

    serialno=0
    clients=0
    meters=0
    protocols=0
    try:
        mbs=os.popen("/binary/mbushub.elf -S")
        serialno=mbs.readline().strip('\n')
        mbs.close()
        mbl=os.popen("/binary/mbushub.elf -l")
        licensearray=mbl.readline().strip('\n').split(';')
        mbl.close()
        clients=licensearray[0]
        meters=licensearray[1]
        protocols=licensearray[2].split(',')
        length = len(licensearray)
        for i in range(length):
            print(licensearray[i])
    except:
        pass

    #Meterlist card
    numberOfMeters = 0
    numberOfWiredMeters = 0
    numberOfRadioMeters = 0
    numberOfNodes = 0
    numberOfUndefinedMeters = 0

    meterlist = []
    try:
        with open("/config/meterlist.csv") as file:
            delimiter = sniffer_csv(file)
            meterlist = list(rcf(file, delimiter))
        meterlist = sorted(meterlist)
    except Exception as e:
        print("error in start_page "+str(e))

    wiredMeters=[]
    radioMeters=[]
    nodes=[]
    undefinedMeters=[]
    for row in meterlist:
        if len(row) > 8:
            if row[8] == "W":
                wiredMeters.append(row)
                numberOfWiredMeters += 1
                numberOfMeters +=1
            elif row[8] == "R":
                radioMeters.append(row)
                numberOfRadioMeters += 1
                numberOfMeters +=1
            elif row[8] == "N":
                nodes.append(row)
                numberOfNodes += 1
                numberOfMeters +=1
            else:
                undefinedMeters.append(row)
                numberOfUndefinedMeters +=1
                numberOfMeters +=1
        else:
            undefinedMeters.append(row)
            numberOfUndefinedMeters +=1
            numberOfMeters +=1

   #Port configuration card

    try:
        mbv=os.popen("/binary/mbushub.elf -V")
        mbusHubVersion=mbv.read()
        mbv.close()
        mbv=os.popen("/binary/mbushub.elf -a")
        licarray=mbv.read().split(',')
        mbv.close()
        nslaves = int(licarray[1])
    except:
        nslaves = 0
        mbusHubVersion = ''
        licarray = ''

    mbushubini_def = configparser.RawConfigParser()
    mbushubini_def.read("/config/mbushub_default.ini")
    mbushubini = configparser.RawConfigParser()
    mbushubini.read("/config/mbushub.ini")

    slaveport = {
        "porttype": str,
        "protocol": str,
        "timeout": int,
        "localport": int,
        "sercomport": str,
        "serbaudrate": int
    }

    try:
        parser = Configuration(mbushubini_def, mbushubini)
        config = parser.read({
            "MasterPort": {
                "porttype": str,
                "protocol": str,
                "timeout": int,
                "sercomport": str,
                "serbaudrate": int,
                "remoteip": str,
                "remoteport": int
            },
            "SlavePort1": slaveport,
            "SlavePort2": slaveport
        })
    except:
        traceback.print_exc()

    if nslaves > 2:
        tmp = parser.read({ "SlavePort3": slaveport })
        config["SlavePort3"] = tmp["SlavePort3"]
        if nslaves > 3:
            tmp = parser.read({ "SlavePort4": slaveport })
            config["SlavePort4"] = tmp["SlavePort4"]

    try:
        quickpostini = configparser.RawConfigParser()
        quickpostini.read("/config/quickpost.ini")
        quickpostPort = quickpostini["Main"]["mbus_ascii_server_port"]
    except:
        quickpostPort = ""
###
    di1=getGPIO("pioA3")
    di2=getGPIO("pioA4")
    relay  = "Undefined"
    try:
        rnc=getGPIO("pioC12")
        rno=getGPIO("pioC10")
        if  (rno == 'On' and rnc == 'Off'):
            relay  = "On"
        elif (rno == 'Off' and rnc == 'On'):
            relay = "Off"
    except:
        pass

    # M-Bus Loop Power
    rs485_failsafe = getGPIO("pioA24")
    loop1 = getGPIO("pioA26")
    loop2 = getGPIO("pioA27")
    loop3 = getGPIO("pioA28")
    loop4 = getGPIO("pioA29")
    # End of M-Bus Loop Power

    board_temperature = "Missing"
    mbus_stage_temperature = "Missing"
    error_flag = 0
    try:
        conf = configparser.RawConfigParser()
        conf.read('/tmp/pi900_state.txt')
        board_temperature = conf.get('Main','Temperature_40V')
        mbus_stage_temperature = conf.get('Main','Temperature_MBus')
        error_flag  = int(conf.get('Main','MBusErrorVal'))
    except:
        pass

    error_message="0, No error"
    if (error_flag & 1):
        error_message = "M-Bus Overload"
    if (error_flag & 2):
        error_message = "M-Bus Short Circuit"
    error_string = "%08X"%error_flag
    mbusCurrent = 0
    mbusMaxCurrent = 0
    try:
        mbusCurrent      = float(conf.get('Main','MbusCurrent'))
        mbusMaxCurrent   = float(conf.get('Main','MbusMaxCurrent'))
    except:
        pass

    mbusCurrentString      = "%.1f mA"%mbusCurrent
    mbusMaxCurrentString   = "%.1f mA"%mbusMaxCurrent
    mbusCurrentPercentUsed = "Unavailable"
    if mbusMaxCurrent > 0:
        mbusCurrentPercentUsed = "%d%%"%int(mbusCurrent/mbusMaxCurrent * 100)

###
    info = {
        'di1'    : di1,
        'di2'    : di2,
        'relay'  : relay,
        'loop1'  : loop1,
        'loop2'  : loop2,
        'loop3'  : loop3,
        'loop4'  : loop4,
        'rs485_failsafe' : rs485_failsafe,
        'board_temperature' : board_temperature,
        'mbus_stage_temperature' : mbus_stage_temperature,
        'error_flag' : error_flag,
        'error_message' : error_message,
        'error_string' : error_string,
        'mbusCurrent' : mbusCurrent,
        'mbusMaxCurrent' : mbusMaxCurrent,
        'mbusCurrentString'      : mbusCurrentString,
        'mbusMaxCurrentString'   : mbusMaxCurrentString,
        'mbusCurrentPercentUsed' : mbusCurrentPercentUsed,
        'mbusHubVersion' : mbusHubVersion,
        'nslaves' : nslaves,
        'numberOfMeters' : numberOfMeters,
        'numberOfWiredMeters' : numberOfWiredMeters,
        'numberOfRadioMeters' : numberOfRadioMeters,
        'numberOfNodes' : numberOfNodes,
        'numberOfUndefinedMeters' : numberOfUndefinedMeters,
        'quickpostPort' : quickpostPort,
        'serialno': serialno,
        'clients': clients,
        'meters': meters,
        'protocols': protocols,
        'DeviceTypes' : DeviceTypes,
        'csvDict' : csvDict,
        'wiredMeters' : wiredMeters,
        'nodes' : nodes,
        'radioMeters' : radioMeters,
        'undefinedMeters' : undefinedMeters,
        'porttypeM' : config["MasterPort"]["porttype"],
        'protocolM' : config["MasterPort"]["protocol"],
        'timeoutM' : config["MasterPort"]["timeout"],
        'sercomportM' : config["MasterPort"]["sercomport"],
        'serbaudrateM' : config["MasterPort"]["serbaudrate"],
        'remoteipM' : config["MasterPort"]["remoteip"],
        'remoteportM' : config["MasterPort"]["remoteport"],
        'porttypeS1' : config["SlavePort1"]["porttype"],
        'protocolS1' : config["SlavePort1"]["protocol"],
        'timeoutS1' : config["SlavePort1"]["timeout"],
        'localportS1' : config["SlavePort1"]["localport"],
        'sercomportS1' : config["SlavePort1"]["sercomport"],
        'serbaudrateS1' : config["SlavePort1"]["serbaudrate"],
        'porttypeS2' : config["SlavePort2"]["porttype"],
        'protocolS2' : config["SlavePort2"]["protocol"],
        'timeoutS2' : config["SlavePort2"]["timeout"],
        'localportS2' : config["SlavePort2"]["localport"],
        'sercomportS2' : config["SlavePort2"]["sercomport"],
        'serbaudrateS2' : config["SlavePort2"]["serbaudrate"],
        'pi900_type' : pi900_type
    }
    if nslaves > 2:
        port3Info = {
            'porttypeS3' : config["SlavePort3"]["porttype"],
            'protocolS3' : config["SlavePort3"]["protocol"],
            'timeoutS3' : config["SlavePort3"]["timeout"],
            'localportS3' : config["SlavePort3"]["localport"],
            'sercomportS3' : config["SlavePort3"]["sercomport"],
            'serbaudrateS3' : config["SlavePort3"]["serbaudrate"]
        }
        info.update(port3Info)
    if nslaves > 3:
        port4Info = {
            'porttypeS4' : config["SlavePort4"]["porttype"],
            'protocolS4' : config["SlavePort4"]["protocol"],
            'timeoutS4' : config["SlavePort4"]["timeout"],
            'localportS4' : config["SlavePort4"]["localport"],
            'sercomportS4' : config["SlavePort4"]["sercomport"],
            'serbaudrateS4' : config["SlavePort4"]["serbaudrate"]
        }
        info.update(port4Info)

    return render_template('start_page.html',  **info)


@app.route("/administration", methods=['GET'])
def administration():
    eth0GW = ''
    eth2GW = ''
    try:
        cf=os.popen("/bin/netstat -r -n")
        gwcs=cf.read().split('\n')
        cf.close()
        for line in gwcs:
            items = line.split()
            if len(items) > 0 and items[0] == '0.0.0.0':
                if items[7] == 'eth0':
                    eth0GW = items[1]
                elif items[7] == 'eth1':
                    eth2GW = items[1]
    except:
        pass

    nameserv=""
    try:
        cf=open("/etc/resolv.conf",'r')
        nwcs=cf.read().rstrip()
        cf.close()
        c=nwcs.split()
        if len(c)>0:
            if c[0]=='nameserver':
                nameserv=c[1]
    except:
        pass

    #CONFIGURATION
    update = request.args.get("UPDATE")
    tab = request.args.get("TAB")

    cf=open("/etc/network/interfaces",'r')
    nwcs=cf.read().split('\n')
    cf.close()

    eth0TYPE=""
    eth2TYPE=""
    eth2IP="None"
    eth2MAC="None"
    eth2SM="None"
    for i in nwcs:
        if not i=='':
            c=i.split()
            if len(c)>0:
                if c[0]=='iface' and c[1]=='eth0':
                    eth0TYPE=c[3]
                if c[0]=='iface' and c[1]=='eth1':
                    eth2TYPE=c[3]
    eth0MAC = ""
    eth0IP=""
    eth0SM=""
    try:
        cf=os.popen("/sbin/ifconfig eth0 2>/dev/null")
        ifcs=cf.read().split('\n')
        cf.close()
        isEth0Static = eth0TYPE == 'static'
        eth0MAC=ifcs[0].split()[4]
        eth0IP=ifcs[1].split()[1].split('addr:')[1]
        eth0SM=ifcs[1].split()[3].split('Mask:')[1]
    except:
        pass

    csvArray=glob.glob("/config/*.[cC][sS][vV]")+glob.glob("/config/*.[xX][mM][lL]")
    for i in range(len(csvArray)):
        csvArray[i] = os.path.basename(csvArray[i])
    iniArray = glob.glob("/config/*.[iI][nN][iI]")
    for i in range(len(iniArray)):
        iniArray[i] = os.path.basename(iniArray[i])

    tzstr=""
    hwtime=""
    ntpIP=""
    try:
        f=open("/etc/TZ",'r')
        tzstr=f.read().strip()
        f.close()
    except:
        pass
    try:
        #tzstr = "UTC+1"
        cf=os.popen("/sbin/hwclock --rtc /dev/rtc1 2>/dev/null")
        hwtime=cf.read().rstrip()
        cf.close()
    except:
        pass
    try:
        f=open("/etc/ntp.conf",'r')
        a=f.read()
        f.close()
        ntpIP=a.split('\n')[0].split()[1]
    except:
        pass

    #UPDATE PAGE
    if update == "TRUE":
        return render_template_string(
            """<HTML>
                    <HEAD>
                        <meta http-equiv=\"refresh\" content=\"0;url=/administration\">"
                    </HEAD>
                </HTML>""")

    randip='Unset'
    if (os.path.isfile('/etc/random_ip')):
        randip='Set'

    inifileArray=glob.glob("/config/*.[iI][nN][iI]")
    for i in range(len(inifileArray)):
        inifileArray[i] = os.path.basename(inifileArray[i])

    securityEnabled = False
    passwordEmpty = True #False if password has been set
    currtime = time.strftime("%Y-%m-%d %H:%M:%S")

    mbushubVersion=""
    quickpostVersion=""
    modbusVersion=""
    wirelessVersion=""
    hanbusVersion=""

    try:
        mbv=os.popen("/binary/mbushub.elf -V")
        mbushubVersion=mbv.read()
        mbv.close()
    except:
        pass
    try:
        qv=os.popen("/binary/quickpost.elf -v")
        quickpostVersion=qv.read()
        qv.close()
    except:
        pass
    try:
        m2m=os.popen("/binary/modbus2mbus -v")
        modbusVersion=m2m.read()
        m2m.close()
    except:
        pass
    try:
        wv=os.popen("/binary/wireless -v")
        wirelessVersion=wv.read()
        wv.close()
    except:
        pass
    try:
        h2m=os.popen("/binary/han2mbus -v")
        hanbusVersion=h2m.read()
        wv.close()
    except:
        pass

    pingIntervalArray = ['Unset','30m','2h','6h','24h']
    cnf = Configs("config","pi900_system")
    ping_address = cnf.getopt('PingReboot','ping_address')
    ping_interval = cnf.getopt('PingReboot','ping_interval')
    if ping_interval not in pingIntervalArray: # This should never happen
        ping_interval='Unset'
    if validIPAddress(ping_address) != "IPv4":
        ping_address = eth0GW

    info = {
        'pingIntervalArray' : pingIntervalArray,
        'ping_address': ping_address,
        'ping_interval': ping_interval,
        'mbushubVersion': mbushubVersion,
        'quickpostVersion': quickpostVersion,
        'modbusVersion': modbusVersion,
        'wirelessVersion': wirelessVersion,
        'hanbusVersion': hanbusVersion,
        'eth0GW': eth0GW,
        'eth2GW': eth2GW,
        'nameserv': nameserv,
        'update': update,
        'tab': tab,
        'eth0TYPE': eth0TYPE,
        'eth2TYPE': eth2TYPE,
        'isEth0Static': isEth0Static,
        'eth0MAC': eth0MAC,
        'eth0IP': eth0IP,
        'eth0SM': eth0SM,
        'eth2IP': eth2IP,
        'eth2MAC': eth2MAC,
        'eth2SM': eth2SM,
        'randip' : randip,
        'csvArray' : csvArray,
        'iniArray' : iniArray,
        'tzstr': tzstr,
        'hwtime': hwtime,
        'ntpIP': ntpIP,
        'inifileArray': inifileArray,
        'securityEnabled': securityEnabled,
        'passwordEmpty': passwordEmpty,
        'currtime': currtime,
        'pi900_type' : pi900_type
    }
    return render_template('administration.html', **info)

@app.route("/licenses", methods=['GET'])
def licenses_route():
    try:
        name = secure_filename(os.path.basename(request.args.get('FILE')))
        return send_file("/var/www1/licenses/" + name)
    except:
        info = { 'license_files': [] }
        for entry in os.listdir("licenses"):
            info['license_files'].append(entry)
        return render_template('licenses.html', **info)


@app.route("/cgi-bin/showFile.py", methods=['GET'])
def showfile():
    cstr=""
    try:
        fname="/config/" + os.path.basename(request.args.get('FILE'))
        f=open(fname,'r')
        cstr=f.read()
        f.close()
    except:
        cstr="No Filename given"

    cstr=cstr.replace('\n','<br>')
    hstr="<HTML>\n<HEAD>\n</HEAD>\n<body>%s</body>\n</HTML>\n"%cstr
    return render_template_string(hstr)

#@app.route("/cgi-bin/basic_settings.py", methods=['GET'])
@app.route("/basic_settings", methods=['GET'])
def base_settings():
    hwconfig = configparser.RawConfigParser()
    hwconfig.read('/config/mbushub_lic.ini')

    serialno=0
    clients=0
    meters=0
    protocols=0
    cards=0
    licensestring=""

    try:
        mbs=os.popen("/binary/mbushub.elf -S")
        serialno=mbs.readline().strip('\n')
        mbs.close()
        mbl=os.popen("/binary/mbushub.elf -l")
        licensearray=mbl.readline().strip('\n').split(';')
        mbl.close()
        clients=licensearray[0]
        meters=licensearray[1]
        protocols=licensearray[2]
        cards=licensearray[3]
        licensestring=hwconfig.get('Main', 'licstr')
    except:
        pass

    https_on=""
    http_on=""
    http_noauth_on=""
    conffile = os.path.basename(os.readlink("/etc/lighttpd/lighttpd.conf"))
    if (conffile == 'lighttpd_https.conf'):
        https_on="checked"
    elif (conffile == 'lighttpd_http.conf'):
        http_on="checked"
    elif (conffile == 'lighttpd_http_noauth.conf'):
        http_noauth_on="checked"
    form = {
        'serialno': serialno,
        'clients': clients,
        'meters': meters,
        'protocols': protocols,
        'cards': cards,
        'licensestring': licensestring,
        'https_on': https_on,
        'http_on': http_on,
        'http_noauth_on': http_noauth_on,
        'conffile': conffile,
    }
    return render_template('basic_settings.html', **form)

def line_prepender(filename, line):
    with open(filename, 'r+') as f:
        content = f.read()
        f.seek(0, 0)
        f.write(line.rstrip('\r\n') + '\n' + content)

@app.route("/downloadCsvFile")
def downloadCsvFile():
    cstr = request.args.get('FILE')
    cstr = secure_filename(cstr)
    cf = []
    temporary = ""
    if cstr == "meterlist.csv":
        try:
            header = "#Id-number,Manufacturer,Version,Medium,Manufacturer(text),Fabrication-number,Primary-address,Verified,Meter-type(Wired/Radio/Wireless Node),Wireless-key,Description,Timestamp"
            line_prepender("/config/meterlist.csv", header)
        except:
            pass

        try:
            return send_file('/config/%s'%cstr, attachment_filename=cstr, mimetype='application/octet-stream', as_attachment=True,  cache_timeout=0)
        except:
            return 'File not found'
    elif cstr == "myconfig.csv" or cstr == "masterpost_Mbus2Modbus.csv" or cstr == "slaveport_Mbus2Modbus.csv":
        with open("/config/"+cstr) as file:
            temporary = file.read()
            with open("/tmp/"+cstr+"_download", "w", encoding="latin-1") as file_write:
                file_write.write(temporary)

        try:
            return send_file('/tmp/%s_download'%cstr, attachment_filename=cstr, mimetype='application/octet-stream', as_attachment=True,  cache_timeout=0)
        except:
            return 'File not found'
    else:
        try:
            return send_file('/config/%s'%cstr, attachment_filename=cstr, mimetype='application/octet-stream', as_attachment=True,  cache_timeout=0)
        except:
            return 'File not found'

@app.route("/copyFromConfig")
def copyFromConfig():
    try:
        os.system("cp /config/meterlist.csv /tmp/")
    except:
        pass
    return redirect('/meter_search_and_overview')

@app.route("/rmCsvFile")
def rmCsvFile():
    try:
        cstr=request.args.get("CSVRMF").replace(' ','\ ')
        filename = secure_filename(cstr)
        os.remove("/config/"+filename)

    except Exception as e:
        print("exception", str(e))
    return render_template_string("""<HTML><HEAD>
            File removed successfully.
            <meta http-equiv=\"refresh\" content=\"2;url=/administration\">
            </HEAD></HTML>""")

@app.route("/cgi-bin/backupConf.py", methods=['POST'])
def backupConf():
    mac = getnode()
    serialno = (mac >> 1) & 0x3FFFFFF
    fname = "pi%sconfig_%d_backup"%(pi900_type,serialno)
    os.system("tar cf /tmp/%s.tar /config\n"%fname)
    os.system("gzip /tmp/%s.tar\n"%fname)
    os.system("mv /tmp/%s.tar.gz /tmp/%s.tgz\n"%(fname,fname))
    return send_file('/tmp/%s.tgz'%fname, attachment_filename='%s.tgz'%fname, mimetype='application/octet-stream', as_attachment=True)

@app.route("/changePassword", methods=['POST'])
def changePassword():
    filename = "/etc/lighttpd/lighttpd.user"
    username = "Admin"

    oldpass=getParameter("pass1")
    npass1=getParameter("pass2")
    npass2=getParameter("pass3")

    def isCorrectOldPassword():
        retval = False
        f = open(filename, "r")
        line = f.readline()
        if line.startswith(username + ":"):
            oldpassword = line[6:len(line)-1]
            if oldpassword == oldpass:
                retval = True
        f.close()
        return retval

    retstr=""
    if  not isCorrectOldPassword():
        retstr = "Old password incorrect."
    elif  npass1 != npass2:
        retstr = "New passwords do not match."
    else:
        f = open(filename, "w")
        f.write(username + ":" + npass1 + "\n")
        retstr = "Password changed successfully."
        f.close()
        os.system("sync")
    return render_template_string("""<HTML><HEAD>
        %s
        <meta http-equiv=\"refresh\" content=\"2;url=/basic_settings\">
        </HEAD></HTML>"""%retstr)


@app.route("/cgi-bin/defaultSettings.py")
def defaultSettings():
    result = """<HTML><HEAD>
            <meta http-equiv=\"refresh\" content=\"2;url=/basic_settings\">
        </HEAD>
        <BODY>"""
    val=""
    if request.args.get("DEFSET"):
        val=request.args.get("DEFSET")

    if val == 'system':
        result += "Setting system configuration to Default"
        os.system("cp -f /default/ntp.conf /etc/\n")
        os.system("cp -f /default/TZ /etc/\n")
    elif val == 'mbushub':
        result += "Setting MBusHub configuration to Default"
        os.system("cp -f /config/mbushub_default.ini /config/mbushub.ini\n")
    elif val == 'quickpost':
        result += "Setting Quickpost configuration to Default"
        os.system("cp -f /config/quickpost_default.ini /config/quickpost.ini\n")
    elif val == 'modbus2mbus':
        result += "Setting Modbus2MBus configuration to Default"
        os.system("cp -f /config/modbus2mbus_default.ini /config/modbus2mbus.ini\n")
    elif val == 'wireless':
        result += "Setting Wireless configuration to Default"
        os.system("cp -f /config/wireless_default.ini /config/wireless.ini\n")
    elif val == 'web':
        result += "Setting web configuration to Default"
        os.system("cp -a /default/lighttpd /etc/\n")
    elif val == 'ip':
        result += "Setting ip configuration to Default"
        os.system("cp -f /default/interfaces /etc/network\n")
        os.system("touch /etc/random_ip\n")
    else:
        result += "Wrong argument, doing nothing\n"
    result += "</BODY></HTML>"
    return render_template_string(result)


@app.route("/cgi-bin/disableSecurity.py")
def disableSecurity():
    #copy disabled security settings file to web servver
    copyfile("/default/lighttpd/conf/lighttpd.conf.no_auth", "/etc/lighttpd.conf")

    #restart web server
    os.system("/etc/lighttpd/restart.sh") # This must be fixed XXXXXXXXXXXXXXXXXXXXXXXXXX

    return render_template_string("""<HTML><HEAD>
        <meta http-equiv=\"refresh\" content=\"0;url=/cgi-bin/getLicense.sh?message=Security+disabled\">
        </HEAD></HTML>""")

@app.route("/cgi-bin/enableSecurity.py")
def enableSecurity():
    return render_template_string("<HTML><HEAD><meta http-equiv=\"refresh\" content=\"0;url=/administration?message=Security+enabled\"></HEAD></HTML>")

@app.route("/cgi-bin/installSoftware.py", methods=['POST'])
def installSoftware():
    success = False
    print("main route")
    # Check if sd-card are mounted on unit
    with open("/proc/mounts") as txt:
        if '/dev/mmcblk0p1' in txt.read():
            print("in dev/mmcblk")
            success = True
    folder = ""
    if success:
        folder = "data"
    else:
        print("installing from /tmp, no sd-card")
        folder = "tmp"
    os.system("mkdir -p /%s/install"%(folder))
    os.system("rm -f /%s/install/install_package.*"%(folder))
    os.system("rm -rf /%s/install/install_package"%folder)
    file = request.files['filename']
    if file:
        file.save("/%s/install/install_package.tgz"%(folder))
        # It's an uploaded file
        # strip leading path from file name to avoid directory traversal attacks
        os.system("cd /%s/install/;gunzip install_package.tgz\n"%(folder))
        os.system("cd /%s/install/;tar xf install_package.tar\n"%(folder))
        os.system("cd /%s/install/install_package; sh install_arm.sh &> /dev/null"%(folder))
        os.system("rm -f /%s/install/install_package.tar\n"%(folder))
        os.system("rm -rf /%s/install/install_package\n"%folder)
    return render_template_string("""<html>
            <head>
                <meta http-equiv=\"refresh\" content=\"4;url=/administration\">
            </head>
            <body class =\"mainBody\">
                <br>
                <h2>The software were successfully installed.</h2>
                <p>Reboot the 900-unit after all installations are completed.</p>
            </body>
        </html>""")

@app.route("/cgi-bin/reboot.py")
def reboot():
    os.system("cp -f /tmp/wireless_telegrams_* /config/tmp/")
    os.system("sync;/sbin/reboot")
    return render_template_string(
        """<html><body>
            System is rebooting.<br />
            Please wait a few minutes and then push the refresh button.
            </body></html>""")


@app.route("/cgi-bin/setClock.py", methods=['POST'])
def setClock():
    hasTimeval=False
    error = False
    timedate="no value"

    if request.form["TimeDate"]:
        hasTimeval = True
        timedate = request.form["TimeDate"].lstrip().split()
        if len(timedate) == 2:
            if len(timedate[0]) == 10 and len(timedate[1])>=8:
                for i in (0,1,2,3,5,6,8,9):
                    if not timedate[0][i].isdigit():
                        error = True
                for i in (0,1,3,4,6,7):
                    if not timedate[1][i].isdigit():
                        error = True
            else:
                error = True
        else:
            error = True
    retstr = ""
    if error:
        retstr = "Incorrect Time Format"
    else:
        if hasTimeval:
            try:
                os.system("date --set=\"%s %s\"\n"%(timedate[0],timedate[1]))
                os.system("hwclock -u -w --rtc /dev/rtc0\n")
                os.system("hwclock -u -w --rtc /dev/rtc1\n")
            except:
                pass
        retstr = "Setting clock, %s"%timedate
    return render_template_string(
        """<html>
            <head>
                <meta http-equiv=\"refresh\" content=\"2;url=/administration\">
            </head>
            <body>
                %s
            </body>
        </html>"""%retstr)

@app.route("/setEthernet",methods=['POST'])
def setEthernet():
    ethernet1Type = request.form["ethernet1Type"]
    if ( ethernet1Type == 'D'): # Dynamic IP
        os.system("cp /etc/network/interfaces_dhcp /tmp/interfaces.tmp")
        if request.form["randomip"] == 'Set':
            os.system("touch /etc/random_ip\n")
        else:
            os.system("rm -f /etc/random_ip\n")
    else:
        ip1 = ""
        sm1 = ""
        try:
            ip1 = request.form['IP1']
            sm1 = request.form['SM1']
        except:
            pass
        
        if validIPAddress(ip1) == "IPv4" and validIPAddress(sm1) == "IPv4":
            stri  = "auto lo\niface lo inet loopback\nauto eth0\n"
            stri += "iface eth0 inet static\n"
            stri += "\taddress %s\n"%ip1
            stri += "\tnetmask %s\n"%sm1
            try:
                gw1=request.form['GW1']
                if validIPAddress(gw1) == "IPv4":
                    stri += "\tgateway %s\n"%gw1
            except:
                pass
            try:
                nserv=request.form['NSERV']
                if nserv == '' or validIPAddress(nserv) != "IPv4":
                    nserv = "8.8.8.8" # Use Google nameserver as default.
                os.system("rm -f /etc/resolv.conf\n") # In case this is a symlink
                with open("/etc/resolv.conf", "w") as f:
                    f.write("nameserver %s\n"%nserv)
            except:
                pass

            with open("/tmp/interfaces.tmp",'w') as tmp:
                tmp.write(stri)

    ping_address = request.form['ping_address']
    ping_interval = request.form['ping_interval']
    if validIPAddress(ping_address) == "IPv4":
        cnf = Configs("config","pi900_system")
        cnf.write_section("PingReboot")
        cnf.save_config()
        stri = ""
        if ping_interval == 'Unset':
            os.system("rm -f /etc/cron_ping_reboot")
        else:
            stri = ""
            if ping_interval == '30m':
                stri = "27,57 * * * *"
            elif ping_interval == '2h':
                stri = "57 */2 * *"
            elif ping_interval == '6h':
                stri = "57 */6 * *"
            elif ping_interval == '24h':
                stri = "57 0 * *"
            if len(stri) > 0:
                stri += " /etc/checkIFStatus.sh eth0 ping %s\n"%ping_address
                with open("/tmp/cron_ping_reboot.tmp",'w') as tmp:
                    tmp.write(stri)
                os.system("mv /tmp/cron_ping_reboot.tmp /etc/cron_ping_reboot; sync\n")

    @after_this_request
    def add_header(response):
        if os.path.isfile("/tmp/interfaces.tmp"):
            os.system("mv /tmp/interfaces.tmp /etc/network/interfaces; sync\n")
            try:
                os.system("reboot")
            except:
                pass
        return response

    return render_template_string("""<HTML>\n<body>
        IP address changed. Type in correct IP-address and refresh browser.
        </body>\n</HTML>""")
            
        

@app.route("/mbushub_setfailsafe", methods=['GET', 'POST'])
def mbushub_SetFailsafe():
    try:
        slnr=int(request.args.get("SLNR"))
    except:
        pass
    urlpart="mbushub_portconf?SLNR=%s"%slnr
    onoff = getParameter("rs485_failsafe")
    setGPIO("pioA24",onoff)
    if (onoff == 'On'):
        os.system("touch /config/rs485_failsafe\n")
    else:
        os.system("rm -f /config/rs485_failsafe\n")
    os.system("sync")

    return render_template_string("""<HTML>\n<HEAD>
            <meta http-equiv=\"refresh\" content=\"2;url=/%s\">
        </HEAD>\n<BODY>
            Setting RS-485 Failsafe
        </BODY>\n</HTML>"""%urlpart)


@app.route("/setIdentification")
def setIdentification():
    #error = False
    identification=""
    if request.args.get("IDENT"):
        try:
            identification = request.args.get("IDENT")
            identification = identification.encode('utf-8')
            identification = identification.decode('utf-8')

        except:
            print ("string is not UTF-8_encode_problem")
            return render_template_string("""<HTML>
            <HEAD>
                <meta http-equiv=\"refresh\" content=\"3;url=/basic_settings\">
            </HEAD>
            <BODY>
                Invalid web interface name. Name can only include utf-8 characters.
            </BODY>
            </HTML>""")
    else:
        print ("Invalid character. String can't start with &")
        return render_template_string("""<HTML>
            <HEAD>
                <meta http-equiv=\"refresh\" content=\"3;url=/basic_settings\">
            </HEAD>
            <BODY>
                Invalid web interface name. Name can only include utf-8 characters.
            </BODY>
            </HTML>""")

    try:
        with open("/var/www1/identification.py", "w") as file:
            file.write(identification)

        return render_template_string("""<HTML>\n<HEAD>
                <meta http-equiv=\"refresh\" content=\"3;url=/cgi-bin/reboot.py\">
            </HEAD>\n<BODY>
                <p>Setting Web Interface identification, {}</p>
            </BODY>\n</HTML>""".format(identification))
    except Exception as e:
        print("error with header.html "+str(e))
        return render_template_string("""<HTML>
            <HEAD>
                <meta http-equiv=\"refresh\" content=\"3;url=/basic_settings\">
            </HEAD>
            <BODY>
                Problem with header.html. Name can only include utf-8 characters.
            </BODY>
            </HTML>""")

@app.route("/restart_server", methods=["GET"])
def restartServer():
    os.system("/etc/init.d/S50lighttpd restart")
    return redirect("/basic_settings")

@app.route("/cgi-bin/setLicense.py", methods=['GET'])
def setLicense():
    lstr=""
    retstr=""
    if request.args.get("licstr"):
        lstr=request.args.get("licstr")
        lstr = lstr.split(" ")[0]
    ret = None
    try:
        p1 = subprocess.Popen(["/binary/mbushub.elf", "-c", lstr],
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            stdin=subprocess.PIPE)
        (p1out,p1err)=p1.communicate()
        print("rc",p1err.decode('utf-8'))
        print("stdout",p1.returncode)
        print("stderr", p1out.decode('utf-8'))
    except Exception as e:
        print("lic err:", str(e))
    if p1.returncode != 0:
        retstr="Incorrect license string. No change made."
    else:
        try:
            cf=open('/config/mbushub_lic.ini','w')
            cf.write("[Main]\n")
            cf.write("licstr=%s\n"%lstr)
            cf.close()
            os.system("/bin/sync")
            os.system("killall mbushub.elf")
            os.system("/binary/mbushub.elf >/dev/null 2>&1 &")
        except:
            pass
        retstr="Setting key to: %s"%lstr
    return render_template_string("""Content-Type: text/html\n
        <HTML>\n<HEAD>
        <meta http-equiv=\"refresh\" content=\"2;url=/basic_settings\">
        </HEAD>\n<BODY>%s</BODY>\n</HTML>"""%retstr)

@app.route("/administration_setntp", methods=['POST'])
def administration_SetNTP():
    error = False
    ntpip=getParameter("ntpIP")

    if ntpip=="":
        os.system("rm -f /etc/default/ntpd\n")
    else:
        f=open("/etc/ntp.conf",'w')
        f.write("""server %s iburst
    server 2.pool.ntp.org iburst
    server 3.pool.ntp.org iburst
    # Allow only time queries, at a limited rate, sending KoD when in excess.
    # Allow all local queries (IPv4, IPv6)
    restrict default nomodify nopeer noquery limited kod
    restrict 127.0.0.1
    restrict [::1]\n"""%ntpip)
        f.flush()
        f.close()
        os.system("sync")
        try:
            os.system("/etc/init.d/S49ntp restart\n")
        except:
            pass
    retstr=""
    if error:
        retstr="Incorrect Address"
    else:
        retstr="Setting NTP to address, %s"%ntpip
    return render_template_string("""<HTML>\n<HEAD>
            <meta http-equiv=\"refresh\" content=\"2;url=/administration\">
        </HEAD>\n<BODY>\n%s\n</BODY>\n</HTML>"""%retstr)

@app.route("/cgi-bin/setTimeZone.py", methods=['POST'])
def setTimeZone():
    error = False
    timezn=getParameter("TimeZone")
    print("timezn", timezn)

    if timezn != "":
        with open("/etc/TZ","w") as file:
            file.write(timezn+"\n")
    else:
        error = True
    retstr = ""
    if error:
        retstr = "Incorrect Time Format"
    else:
        retstr = "Setting Timezone"
    return render_template_string(
        """<HTML>\n<HEAD>\n
                <meta http-equiv=\"refresh\" content=\"2;url=/administration\">
            </HEAD>\n<BODY>\n%s, %s\n</BODY>\n</HTML>"""%(retstr,timezn))

@app.route("/cgi-bin/setWebConf.py", methods=['POST'])
def setWebConf():
    https_val=getParameter("https")

    if (https_val == "https" or https_val == "http" or https_val == "http_noauth"):
        os.system("rm -f /etc/lighttpd/lighttpd.conf\n")
        if https_val == "http":
            os.system("ln -s /etc/lighttpd/lighttpd_http.conf /etc/lighttpd/lighttpd.conf\n")
        elif https_val == "http_noauth":
            os.system("ln -s /etc/lighttpd/lighttpd_http_noauth.conf /etc/lighttpd/lighttpd.conf\n")
        else:
            os.system("ln -s /etc/lighttpd/lighttpd_https.conf /etc/lighttpd/lighttpd.conf\n")
        try:
            os.system("/etc/init.d/S50lighttpd restart &>/dev/null\n")
            os.system("sync")
        except:
            pass
    return render_template_string("""<HTML>\n<HEAD>
            <meta http-equiv=\"refresh\" content=\"2;url=/basic_settings\">
            </HEAD>\n<BODY>
                Setting https, %s
            </BODY>\n</HTML>"""%https_val)

@app.route("/cgi-bin/showFile.py")
def showFile():
    cstr=""
    try:
        fname="/config/" + request.args.get('FILE')
        f=open(fname,'r')
        cstr=f.read()
        f.close()
    except:
        cstr="No Filename given"
    cstr=cstr.replace('\n','<br>')
    hstr="<HTML>\n<HEAD>\n</HEAD>\n<body>%s</body>\n</HTML>\n"%cstr
    return render_template_string(hstr)

@app.route("/cgi-bin/showCsvFile.py")
def showCsvFile():
    str0 = "\n\n<html>\n<head>\n<meta http-equiv=\"Content-Type\" content=\"text/html; charset=UTF-8\"/>\n</head>\n<body>\n<table border=1>\n<tr>\n<td>"
    try:
        fname='/config/'+os.path.basename(request.args.get('FILE'))
        with open(fname,mode='r',encoding='utf-8',errors='replace') as file:
            a = file.read()
            for i in range(0,37):
                str0 += "%d</td><td>"%i
            str0 += "</td>\n</tr>\n<tr>\n<td>"
            str0 += "%s"%a.replace(',','</td><td>').replace('\r','').replace('\n','</td>\n</tr>\n<tr>\n<td>').replace('<td></td>','<td>&nbsp;</td>')
            str0 += "</td>\n</tr>\n</body>\n</html>"
            return render_template_string(str0)
        
    except Exception as e:
        str0 += "<h1>Failed to show csv file. </h1> </br> <p>" + str(e) + "</p>"
        return render_template_string(str0)

@app.route("/cgi-bin/showBrowseCsvFile.py")
def showBrowseCsvFile():
    fname='/tmp/'+os.path.basename(request.args.get('FILE'))
    f=open(fname,encoding="utf-8")
    a=f.read()
    f.close()
    str0 = "\n\n<html>\n<head>\n<meta charset=\"UTF-8\">\n</head>\n<body>\n<table border=1>\n<tr>\n<td>"
    for i in range(0,37):
        str0 += "%d</td><td>"%i
    str0 += "</td>\n</tr>\n<tr>\n<td>"
    str0 += "%s"%a.replace(',','</td><td>').replace('\r','').replace('\n','</td>\n</tr>\n<tr>\n<td>').replace('<td></td>','<td>&nbsp;</td>')
    str0 += "</td>\n</tr>\n</body>\n</html>"
    return render_template_string(str0)


@app.route("/cgi-bin/uploadBackupConf.py", methods=['POST'])
def uploadBackupConf():
    fil = request.files['bconffile']
    # if user does not select file, browser also
    # submit an empty part without filename
    #if file.filename == '':
    #    flash('No selected file')
    #    return redirect(request.url)
    if fil:
        fil.save("/tmp/pi%sconfig_backup.tgz"%pi900_type)
        os.system("cd /tmp/ && gunzip pi%sconfig_backup.tgz"%pi900_type)
        os.system("cd /tmp/ && tar xf pi%sconfig_backup.tar"%pi900_type)
        os.system("rm -f /tmp/config/mbushub_lic.ini") # Do not overwrite the license
        os.system("rm -f /tmp/config/*_default.ini") # Do not overwrite any _default.ini file
        os.system("cp -f /tmp/config/*.ini /config/")
        os.system("cd /tmp/config/ && cp -f *.[cC][sS][vV] /config/\n")
        os.system("cd /tmp/config/ && cp -f *.[xX][mM][lL] /config/\n")
        os.system("cp -a /tmp/config/ca-certificates/* /config/ca-certificates")
        os.system("cp -a /tmp/config/ssh/* /config/ssh")
        os.system("rm -rf /tmp/config/ rm -f pi%sconfig_backup.tar\n"%pi900_type)
    return render_template_string("""<HTML><HEAD>
            <meta http-equiv=\"refresh\" content=\"0;url=/administration\">
            </HEAD></HTML>""")

@app.route("/uploadCSV", methods=['GET','POST'])
def uploadCSV():
    file = request.files['conffile']
    if file and allowed_file(file.filename):
        filename = secure_filename(file.filename)
        file.save(os.path.join("/config", filename))
    return render_template_string("""<HTML><HEAD>
            <meta http-equiv=\"refresh\" content=\"0;url=/administration\">
            </HEAD></HTML>""")

@app.route("/cgi-bin/uploadCSVToOverview.py", methods=['GET','POST'])
def uploadCSVToOverview():
    file = request.files['conffile']
    if file and allowed_file(file.filename):
        filename = secure_filename(file.filename)
        file.save(os.path.join("/config", filename))
    return render_template_string("""<HTML><HEAD>
            <meta http-equiv=\"refresh\" content=\"0;url=/cgi-bin/overview.py\">
            </HEAD></HTML>""")



if __name__ == "__main__":
    if os.environ['FLASK_ENV'] == 'development':
        app.run(use_debugger=False, debug=True, use_reloader=False, host='0.0.0.0',threaded=True)
